"""Load Synthetic dataset Metrics
"""
import json

import dask.bag as db

from datasetinsights.constants import DEFAULT_DATA_ROOT

from .exceptions import DefinitionIDError
from .tables import DATASET_TABLES, SCHEMA_VERSION, glob
from .validation import verify_version


class Metrics:
    """Load metrics table

    Metrics store extra metadata that can be used to describe a particular
    sequence, capture or annotation. Metric records are stored as arbitrary
    number (M) of key-value pairs.
    For more detail, see schema design doc:
    :ref:`metrics`

    Attributes:
        metrics (dask.bag.core.Bag): a collection of metrics records
    Examples:
        >>> metrics = Metrics(data_root="/data")
        >>> metrics_df = metrics.filter_metrics(def_id="my_definition_id")
        #metrics_df now contains all the metrics data corresponding to
        "my_definition_id"

        One example of metrics_df (first row shown below):

        +---------------+------------------+---------------------+
        | label_id(int) | instance_id(int) | visible_pixels(int) |
        +===============+==================+=====================+
        | 2             | 2                | 2231                |
        +---------------+------------------+---------------------+

    """

    TABLE_NAME = "metrics"
    FILE_PATTERN = DATASET_TABLES[TABLE_NAME].file

    def __init__(self, data_root=DEFAULT_DATA_ROOT, version=SCHEMA_VERSION):
        """ Initialize Metrics

        Args:
            data_root (str): the root directory of the dataset containing
            metrics
            version (str): desired schema version
        """
        self.metrics = self._load_metrics(data_root, version)

    def _load_metrics(self, data_root, version):
        """
        `:ref:`metrics`


        Args:
            data_root: (str): the root directory of the dataset containing
            metrics
            version (str): desired schema version

        Returns:
            dask.bag.core.Bag
        """
        metrics_files = db.from_sequence(glob(data_root, self.FILE_PATTERN))
        metrics = metrics_files.map(
            lambda path: Metrics._load_json(path, self.TABLE_NAME, version)
        ).flatten()

        return metrics

    @staticmethod
    def _normalize_values(metric):
        """ Filter unnecessary info from metric.
        1-level faltten of metrics.values column.
        """
        values = metric["values"]
        for value in values:
            value["capture_id"] = metric["capture_id"]
            value["annotation_id"] = metric["annotation_id"]
            value["sequence_id"] = metric["sequence_id"]
            value["step"] = metric["step"]

        return values

    def filter_metrics(self, def_id):
        """Get all metrics filtered by a given metric definition id

        Args:
            def_id (str): metric definition id used to filter results
        Raises:
            DefinitionIDError: raised if no metrics records match the given
            def_id
        Returns (pd.DataFrame):
        Columns: "label_id", "capture_id", "annotation_id", "sequence_id",
        "step"
        """
        metrics = (
            self.metrics.filter(
                lambda metric: metric["metric_definition"] == def_id
            )
            .map(Metrics._normalize_values)
            .flatten()
        )
        if metrics.count().compute() == 0:
            msg = (
                f"Can't find metrics records associated with the given "
                f"definition id {def_id}."
            )
            raise DefinitionIDError(msg)

        return metrics.to_dataframe().compute()

    @staticmethod
    def _load_json(filename, table_name, version):
        """Load records from json files into a dict
        """
        with open(filename, "r") as file:
            data = json.load(file)
        verify_version(data, version)

        return data[table_name]



""" Simulation Dataset Catalog
"""


import logging

from pyquaternion import Quaternion

from datasetinsights.io.bbox import BBox2D, BBox3D

logger = logging.getLogger(__name__)


def read_bounding_box_3d(annotation, label_mappings=None):
    """ Convert dictionary representations of 3d bounding boxes into objects
    of the BBox3d class

    Args:
        annotation (List[dict]): 3D bounding box annotation
        label_mappings (dict): a dict of {label_id: label_name} mapping

    Returns:
        A list of 3d bounding box objects
    """

    bboxes = []

    for b in annotation:
        label_id = b["label_id"]
        translation = (
            b["translation"]["x"],
            b["translation"]["y"],
            b["translation"]["z"],
        )
        size = (b["size"]["x"], b["size"]["y"], b["size"]["z"])
        rotation = b["rotation"]
        rotation = Quaternion(
            x=rotation["x"], y=rotation["y"], z=rotation["z"], w=rotation["w"]
        )

        if label_mappings and label_id not in label_mappings:
            continue
        box = BBox3D(
            translation=translation,
            size=size,
            label=label_id,
            sample_token=0,
            score=1,
            rotation=rotation,
        )
        bboxes.append(box)

    return bboxes


def read_bounding_box_2d(annotation, label_mappings=None):
    """Convert dictionary representations of 2d bounding boxes into objects
    of the BBox2D class

    Args:
        annotation (List[dict]): 2D bounding box annotation
        label_mappings (dict): a dict of {label_id: label_name} mapping

    Returns:
        A list of 2D bounding box objects
    """
    bboxes = []
    for b in annotation:
        label_id = b["label_id"]
        x = b["x"]
        y = b["y"]
        w = b["width"]
        h = b["height"]
        if label_mappings and label_id not in label_mappings:
            continue
        box = BBox2D(label=label_id, x=x, y=y, w=w, h=h)
        bboxes.append(box)

    return bboxes



import logging
import os

from datasetinsights.io.download import (
    download_file,
    get_checksum_from_file,
    validate_checksum,
)
from datasetinsights.io.downloader.base import DatasetDownloader
from datasetinsights.io.exceptions import ChecksumError

logger = logging.getLogger(__name__)


class HTTPDatasetDownloader(DatasetDownloader, protocol="http://"):
    """ This class is used to download data from any HTTP or HTTPS public url
        and perform function such as downloading the dataset and checksum
        validation if checksum file path is provided.
    """

    def download(self, source_uri, output, checksum_file=None, **kwargs):
        """ This method is used to download the dataset from HTTP or HTTPS url.

        Args:
            source_uri (str): This is the downloader-uri that indicates where
                              the dataset should be downloaded from.

            output (str): This is the path to the directory where the download
                          will store the dataset.

            checksum_file (str): This is path of the txt file that contains
                                 checksum of the dataset to be downloaded. It
                                 can be HTTP or HTTPS url or local path.

        Raises:
            ChecksumError: This will raise this error if checksum doesn't
                           matches

        """
        dataset_path = download_file(source_uri, output)

        if checksum_file:
            logger.debug("Reading checksum from checksum file.")
            checksum = get_checksum_from_file(checksum_file)
            try:
                logger.debug("Validating checksum!!")
                validate_checksum(dataset_path, int(checksum))
            except ChecksumError as e:
                logger.info("Checksum mismatch. Deleting the downloaded file.")
                os.remove(dataset_path)
                raise e



MAX_SAMPLES = 10000
RENDERED_OBJECT_INFO_DEFINITION_ID = "659c6e36-f9f8-4dd6-9651-4a80e51eabc4"
USER_PARAMETERS_DEFINITION_ID = "3f06bcec-1f23-4387-a1fd-5af54ee29c16"
FOREGROUND_PLACEMENT_INFO_DEFINITION_ID = "061e08cc-4428-4926-9933-a6732524b52b"
LIGHTING_INFO_DEFINITION_ID = "939248ee-668a-4e98-8e79-e7909f034a47"
BOUNDING_BOX_2D_DEFINITION_ID = "c31620e3-55ff-4af6-ae86-884aa0daa9b2"



"""UnitySimulationDownloader downloads a dataset from Unity Simulation"""
import concurrent.futures
import logging
import os
import re
from pathlib import Path

import numpy as np
import pandas as pd
import requests
from codetiming import Timer
from requests.packages.urllib3.util.retry import Retry
from tqdm import tqdm

import datasetinsights.constants as const
from datasetinsights.datasets.unity_perception.tables import (
    DATASET_TABLES,
    FileType,
)
from datasetinsights.io.download import TimeoutHTTPAdapter, download_file
from datasetinsights.io.downloader.base import DatasetDownloader
from datasetinsights.io.exceptions import DownloadError

# number of workers for ThreadPoolExecutor. This is the default value
# in python3.8
MAX_WORKER = min(32, os.cpu_count() + 4)
# Timeout of requests (in seconds)
DEFAULT_TIMEOUT = 1800
# Retry after failed request
DEFAULT_MAX_RETRIES = 5


logger = logging.getLogger(__name__)


class UnitySimulationDownloader(DatasetDownloader, protocol="usim://"):
    """ This class is used to download data from Unity Simulation

        For more on Unity Simulation please see these
        `docs <https://github.com/Unity-Technologies/Unity-Simulation-Docs>`

        Args:
            access_token (str): Access token to be used to authenticate to
                unity simulation for downloading the dataset

    """

    SOURCE_URI_PATTERN = r"usim://([^@]*)?@?([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})/(\w+)"  # noqa: E501

    def __init__(self, access_token=None, **kwargs):
        super().__init__(**kwargs)
        self.access_token = access_token
        self.run_execution_id = None
        self.project_id = None

    def download(self, source_uri, output, include_binary=False, **kwargs):
        """ Download from Unity Simulation

        Args:
            source_uri: This is the downloader-uri that indicates where on
                unity simulation the dataset should be downloaded from.
                The expected source-uri should follow these patterns:
                usim://access-token@project-id/run-execution-id
                or
                usim://project-id/run-execution-id
            output: This is the path to the directory where the download
                method will store the dataset.
            include_binary: Whether to download binary files such as images
                or LIDAR point clouds. This flag applies to Datasets where
                metadata (e.g. annotation json, dataset catalog, ...)
                can be separated from binary files.

        """
        self.parse_source_uri(source_uri)
        manifest_file = os.path.join(output, f"{self.run_execution_id}.csv")
        manifest_file = download_manifest(
            self.run_execution_id,
            manifest_file,
            self.access_token,
            project_id=self.project_id,
        )

        dl_worker = Downloader(manifest_file, output)
        dl_worker.download_references()
        dl_worker.download_metrics()
        dl_worker.download_captures()
        if include_binary:
            dl_worker.download_binary_files()

    def parse_source_uri(self, source_uri):
        """ Parse unity simulation source uri

        Args:
            source_uri: Parses source-uri in the following format
            usim://access-token@project-id/run-execution-id
            or
            usim://project-id/run-execution-id

        """
        pattern = re.compile(self.SOURCE_URI_PATTERN)
        result = pattern.findall(source_uri)
        if len(result) == 1:
            (access_token, project_id, run_execution_id,) = pattern.findall(
                source_uri
            )[0]
            if not self.access_token:
                if access_token:
                    self.access_token = access_token
                else:
                    raise ValueError(f"Missing access token")
            if project_id:
                self.project_id = project_id
            if run_execution_id:
                self.run_execution_id = run_execution_id

        else:
            raise ValueError(
                f"{source_uri} needs to be in format"
                f" usim://access_token@project_id/run_execution_id "
                f"or usim://project_id/run_execution_id "
            )


def _filter_unsuccessful_attempts(manifest_df):
    """
    remove all rows from a dataframe where a greater attempt_id exists for
    the 'instance_id'. This is necessary so that we avoid using data from
    a failed USim run and only get the most recent retry.
    Args:
        manifest_df (pandas df): must have columns 'attempt_id', 'app_param_id'
        and 'instance_id'

    Returns(pandas df): where all rows for earlier attempt ids have been
    removed

    """
    last_attempt_per_instance = manifest_df.groupby("instance_id")[
        "attempt_id"
    ].agg(["max"])
    merged = manifest_df.merge(
        how="outer",
        right=last_attempt_per_instance,
        left_on="instance_id",
        right_on="instance_id",
    )
    filtered = merged[merged["attempt_id"] == merged["max"]]
    filtered = filtered.reset_index(drop=True)
    filtered = filtered.drop(columns="max")
    return filtered


class Downloader:
    """Parse a given manifest file to download simulation output

    For more on Unity Simulation please see these
    `docs <https://github.com/Unity-Technologies/Unity-Simulation-Docs>`_

    Attributes:
        manifest (DataFrame): the csv manifest file stored in a pandas dataframe
        data_root (str): root directory where the simulation output should
            be downloaded
    """

    MANIFEST_FILE_COLUMNS = (
        "run_execution_id",
        "app_param_id",
        "instance_id",
        "attempt_id",
        "file_name",
        "download_uri",
    )

    def __init__(self, manifest_file: str, data_root: str):
        """ Initialize Downloader

        Args:
            manifest_file (str): path to a manifest file
            data_root (str): root directory where the simulation output should
                be downloaded
        """
        self.manifest = pd.read_csv(
            manifest_file, header=0, names=self.MANIFEST_FILE_COLUMNS
        )
        self.manifest = _filter_unsuccessful_attempts(manifest_df=self.manifest)
        self.manifest["filetype"] = self.match_filetypes(self.manifest)
        self.data_root = data_root

    @staticmethod
    def match_filetypes(manifest):
        """ Match filetypes for every rows in the manifest file.

        Args:
            manifest (pd.DataFrame): the manifest csv file

        Returns:
            a list of filetype strings
        """
        filenames = manifest.file_name
        filetypes = []
        for name in filenames:
            for _, table in DATASET_TABLES.items():
                if re.match(table.pattern, name):
                    filetypes.append(table.filetype)
                    break
            else:
                filetypes.append(FileType.BINARY)

        return filetypes

    @Timer(name="download_all", text=const.TIMING_TEXT, logger=logging.info)
    def download_all(self):
        """ Download all files in the manifest file.
        """
        matched_rows = np.ones(len(self.manifest), dtype=bool)
        downloaded = self._download_rows(matched_rows)
        logger.info(
            f"Total {len(downloaded)} files in manifest are successfully "
            f"downloaded."
        )

    @Timer(
        name="download_references", text=const.TIMING_TEXT, logger=logging.info
    )
    def download_references(self):
        """ Download all reference files.
        All reference tables are static tables during the simulation.
        This typically comes from the definition of the simulation and should
        be created before tasks running distributed at different instances.
        """
        logger.info("Downloading references files...")
        matched_rows = self.manifest.filetype == FileType.REFERENCE
        downloaded = self._download_rows(matched_rows)

        logger.info(
            f"Total {len(downloaded)} reference files are successfully "
            f"downloaded."
        )

    @Timer(name="download_metrics", text=const.TIMING_TEXT, logger=logging.info)
    def download_metrics(self):
        """ Download all metrics files.
        """
        logger.info("Downloading metrics files...")
        matched_rows = self.manifest.filetype == FileType.METRIC
        downloaded = self._download_rows(matched_rows)
        logger.info(
            f"Total {len(downloaded)} metric files are successfully downloaded."
        )

    @Timer(
        name="download_captures", text=const.TIMING_TEXT, logger=logging.info
    )
    def download_captures(self):
        """ Download all captures files. See :ref:`captures`
        """
        logger.info("Downloading captures files...")
        matched_rows = self.manifest.filetype == FileType.CAPTURE
        downloaded = self._download_rows(matched_rows)
        logger.info(
            f"Total {len(downloaded)} capture files are successfully "
            f"downloaded."
        )

    @Timer(
        name="download_binary_files",
        text=const.TIMING_TEXT,
        logger=logging.info,
    )
    def download_binary_files(self):
        """ Download all binary files.
        """
        logger.info("Downloading binary files...")
        matched_rows = self.manifest.filetype == FileType.BINARY
        downloaded = self._download_rows(matched_rows)
        logger.info(
            f"Total {len(downloaded)} binary files are successfully "
            f"downloaded."
        )

    def _download_rows(self, matched_rows):
        """ Download matched rows in a manifest file.

        Note:
        We might need to download 1M+ of simulation output files, in this case
        we don't want to have a single file transfer failure holding back on
        getting the simulation data. Here download exception are captured.
        We only log an error message and requires uses to pay attention to
        this error.

        Args:
            matched_rows (pd.Series): boolean series indicator of the manifest
                file that should be downloaded

        Returns:
            list of strings representing the downloaded destination path.
        """
        n_expected = sum(matched_rows)
        future_downloaded = []
        downloaded = []
        with concurrent.futures.ThreadPoolExecutor(MAX_WORKER) as executor:
            for _, row in self.manifest[matched_rows].iterrows():
                source_uri = row.download_uri
                relative_path = Path(self.data_root, row.file_name)
                dest_path = relative_path.parent
                file_name = relative_path.name
                future = executor.submit(
                    download_file, source_uri, dest_path, file_name
                )
                future_downloaded.append(future)

            for future in tqdm(
                concurrent.futures.as_completed(future_downloaded),
                total=n_expected,
            ):
                try:
                    downloaded.append(future.result())
                except DownloadError as ex:
                    logger.error(ex)

        n_downloaded = len(downloaded)
        if n_downloaded != n_expected:
            logger.warning(
                f"Found {n_expected} matching records in the manifest file, "
                f"but only {n_downloaded} are downloaded."
            )

        return downloaded


def download_manifest(
    run_execution_id, manifest_file, access_token, project_id, use_cache=True
):
    """ Download manifest file from a single run_execution_id
    For more on Unity Simulation see these
    `docs <https://github.com/Unity-Technologies/Unity-Simulation-Docs>`_


    Args:
        run_execution_id (str): Unity Simulation run execution id
        manifest_file (str): path to the destination of the manifest_file
        access_token (str): short lived authorization token
        project_id (str): Unity project id that has Unity Simulation enabled
        use_cache (bool, optional): indicator to skip download if manifest
            file already exists. Default: True.

    Returns:
        str: Full path to the manifest_file
    """
    api_endpoint = const.USIM_API_ENDPOINT
    project_url = f"{api_endpoint}/v1/projects/{project_id}/"
    data_url = f"{project_url}runs/{run_execution_id}/data"
    if Path(manifest_file).exists() and use_cache:
        logger.info(
            f"Mainfest file {manifest_file} already exists. Skipping downloads."
        )
        return manifest_file

    logger.info(
        f"Trying to download manifest file for run-execution-id "
        f"{run_execution_id}"
    )
    adapter = TimeoutHTTPAdapter(
        timeout=DEFAULT_TIMEOUT, max_retries=Retry(total=DEFAULT_MAX_RETRIES)
    )
    headers = {
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/json",
    }
    with requests.Session() as http:
        http.mount("https://", adapter)
        try:
            resp = http.get(data_url, headers=headers)
            resp.raise_for_status()
        except requests.exceptions.RequestException as ex:
            logger.error(ex)
            err_msg = (
                f"Failed to download manifest file for run-execution-id: "
                f"{run_execution_id}."
            )
            raise DownloadError(err_msg)
        else:
            Path(manifest_file).parent.mkdir(parents=True, exist_ok=True)
            with open(manifest_file, "wb") as f:
                for chunk in resp.iter_content(chunk_size=1024):
                    f.write(chunk)

    logger.info(
        f"Manifest file {manifest_file} downloaded for run-execution-id "
        f"{run_execution_id}"
    )

    return manifest_file



from datasetinsights.io.downloader.base import DatasetDownloader
from datasetinsights.io.gcs import GCSClient


class GCSDatasetDownloader(DatasetDownloader, protocol="gs://"):
    """ This class is used to download data from GCS
    """

    def __init__(self, **kwargs):
        """ initiating GCSDownloader
        """
        self.client = GCSClient()

    def download(self, source_uri=None, output=None, **kwargs):
        """

        Args:
            source_uri: This is the downloader-uri that indicates where on
                GCS the dataset should be downloaded from.
                The expected source-uri follows these patterns
                gs://bucket/folder or gs://bucket/folder/data.zip

            output: This is the path to the directory
                where the download will store the dataset.
        """
        self.client.download(local_path=output, url=source_uri)



import json

import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output

import datasetinsights.stats.statistics as stat
import datasetinsights.stats.visualization.constants as constants

from .app import get_app
from .plots import bar_plot, histogram_plot

app = get_app()


def generate_total_counts_figure(max_samples, roinfo):
    """ Method for generating total object count bar plot using ploty.

    Args:
        max_samples(int): maximum number of samples that will be included
            in the plot.
        roinfo(datasetinsights.data.datasets.statistics.RenderedObjectInfo):
            Rendered Object Info in Captures.

    Returns:
        plotly.graph_objects.Figure: chart to display total object count
    """

    total_counts_fig = bar_plot(
        roinfo.total_counts(),
        x="label_id",
        y="count",
        x_title="Label Id",
        y_title="Count",
        title="Total Object Count in Dataset",
        hover_name="label_name",
    )
    return total_counts_fig


def generate_per_capture_count_figure(max_samples, roinfo):
    """ Method for generating object count per capture histogram using ploty.

    Args:
        max_samples(int): maximum number of samples that will be included
            in the plot.
        roinfo(datasetinsights.data.datasets.statistics.RenderedObjectInfo):
            Rendered Object Info in Captures.

    Returns:
        plotly.graph_objects.Figure: chart to display object counts per capture
    """

    per_capture_count_fig = histogram_plot(
        roinfo.per_capture_counts(),
        x="count",
        x_title="Object Counts Per Capture",
        y_title="Frequency",
        title="Distribution of Object Counts Per Capture: Overall",
        max_samples=max_samples,
    )
    return per_capture_count_fig


def generate_pixels_visible_per_object_figure(max_samples, roinfo):
    """ Method for generating pixels visible per object histogram using ploty.

    Args:
        max_samples(int): maximum number of samples that will be included
            in the plot.
        roinfo(datasetinsights.data.datasets.statistics.RenderedObjectInfo):
            Rendered Object Info in Captures.

    Returns:
        plotly.graph_objects.Figure: chart to display visible pixels per object
    """

    pixels_visible_per_object_fig = histogram_plot(
        roinfo.raw_table,
        x="visible_pixels",
        x_title="Visible Pixels Per Object",
        y_title="Frequency",
        title="Distribution of Visible Pixels Per Object: Overall",
        max_samples=max_samples,
    )

    return pixels_visible_per_object_fig


def html_overview(data_root):
    """ Method for displaying overview statistics.

    Args:
        data_root(str): path to the dataset.

    Returns:
        html layout: displays graphs for overview statistics.
    """

    roinfo = stat.RenderedObjectInfo(
        data_root=data_root, def_id=constants.RENDERED_OBJECT_INFO_DEFINITION_ID
    )
    label_names = roinfo.total_counts()["label_name"].unique()

    total_counts_fig = generate_total_counts_figure(
        constants.MAX_SAMPLES, roinfo
    )
    per_capture_count_fig = generate_per_capture_count_figure(
        constants.MAX_SAMPLES, roinfo
    )
    pixels_visible_per_object_fig = generate_pixels_visible_per_object_figure(
        constants.MAX_SAMPLES, roinfo
    )

    overview_layout = html.Div(
        [
            html.Div(id="overview"),
            dcc.Markdown(
                """ # Total Object Count  """, style={"text-align": "center"}
            ),
            dcc.Graph(id="total_count", figure=total_counts_fig,),
            html.Div(
                [
                    dcc.Markdown(
                        """ # Object Count Distribution """,
                        style={"text-align": "center"},
                    ),
                    dcc.Dropdown(
                        id="object_count_filter",
                        options=[{"label": i, "value": i} for i in label_names],
                        value=label_names[0],
                    ),
                ],
            ),
            html.Div(
                [
                    dcc.Graph(
                        id="per_object_count", figure=per_capture_count_fig,
                    ),
                    dcc.Graph(id="per_object_count_filter_graph",),
                ],
                style={"columnCount": 2},
            ),
            html.Div(
                [
                    dcc.Markdown(
                        """# Visible Pixels Distribution """,
                        style={"text-align": "center"},
                    ),
                    dcc.Dropdown(
                        id="pixels_visible_filter",
                        options=[{"label": i, "value": i} for i in label_names],
                        value=label_names[0],
                    ),
                ],
            ),
            html.Div(
                [
                    dcc.Graph(
                        id="pixels_visible_per_object",
                        figure=pixels_visible_per_object_fig,
                    ),
                    dcc.Graph(id="pixels_visible_filter_graph",),
                ],
                style={"columnCount": 2},
            ),
        ],
    )
    return overview_layout


@app.callback(
    Output("pixels_visible_filter_graph", "figure"),
    [
        Input("pixels_visible_filter", "value"),
        Input("data_root_value", "children"),
    ],
)
def update_visible_pixels_figure(label_value, json_data_root):
    """ Method for generating pixels visible histogram for selected object.
    Args:
        label_value (str): value selected by user using drop-down
    Returns:
        plotly.graph_objects.Figure: displays visible pixels distribution.
    """
    roinfo = stat.RenderedObjectInfo(
        data_root=json.loads(json_data_root),
        def_id=constants.RENDERED_OBJECT_INFO_DEFINITION_ID,
    )
    filtered_roinfo = roinfo.raw_table[
        roinfo.raw_table["label_name"] == label_value
    ][["visible_pixels"]]
    filtered_figure = histogram_plot(
        filtered_roinfo,
        x="visible_pixels",
        x_title="Visible Pixels For " + str(label_value),
        y_title="Frequency",
        title="Distribution of Visible Pixels For " + str(label_value),
        max_samples=constants.MAX_SAMPLES,
    )
    return filtered_figure


@app.callback(
    Output("per_object_count_filter_graph", "figure"),
    [
        Input("object_count_filter", "value"),
        Input("data_root_value", "children"),
    ],
)
def update_object_counts_capture_figure(label_value, json_data_root):
    """ Method for generating object count per capture histogram for selected
        object.
    Args:
        label_value (str): value selected by user using drop-down
    Returns:
        plotly.graph_objects.Figure: displays object count distribution.
    """
    roinfo = stat.RenderedObjectInfo(
        data_root=json.loads(json_data_root),
        def_id=constants.RENDERED_OBJECT_INFO_DEFINITION_ID,
    )
    filtered_object_count = roinfo.raw_table[
        roinfo.raw_table["label_name"] == label_value
    ]
    filtered_object_count = (
        filtered_object_count.groupby(["capture_id"])
        .size()
        .to_frame(name="count")
        .reset_index()
    )
    filtered_figure = histogram_plot(
        filtered_object_count,
        x="count",
        x_title="Object Counts Per Capture For " + str(label_value),
        y_title="Frequency",
        title="Distribution of Object Counts Per Capture For "
        + str(label_value),
        max_samples=constants.MAX_SAMPLES,
    )
    return filtered_figure



import json
import logging
import pathlib
from collections import namedtuple
from enum import Enum

import pandas as pd

from .validation import verify_version

logger = logging.getLogger(__name__)
SCHEMA_VERSION = "0.0.1"  # Synthetic dataset schema version


class FileType(Enum):
    BINARY = "binary"
    REFERENCE = "reference"
    METRIC = "metric"
    CAPTURE = "capture"


Table = namedtuple("Table", "file pattern filetype")
DATASET_TABLES = {
    "annotation_definitions": Table(
        "**/annotation_definitions.json",
        r"(?:\w|-|/)*annotation_definitions.json",
        FileType.REFERENCE,
    ),
    "captures": Table(
        "**/captures_*.json",
        r"(?:\w|-|/)*captures_[0-9]+.json",
        FileType.CAPTURE,
    ),
    "egos": Table("**/egos.json", r"(?:\w|-|/)*egos.json", FileType.REFERENCE),
    "metric_definitions": Table(
        "**/metric_definitions.json",
        r"(?:\w|-|/)*metric_definitions.json",
        FileType.REFERENCE,
    ),
    "metrics": Table(
        "**/metrics_*.json", r"(?:\w|-|/)*metrics_[0-9]+.json", FileType.METRIC
    ),
    "sensors": Table(
        "**/sensors.json", r"(?:\w|-|/)*sensors.json", FileType.REFERENCE
    ),
}


def glob(data_root, pattern):
    """Find all matching files in a directory.

    Args:
        data_root (str): directory containing capture files
        pattern (str): Unix file pattern

    Yields:
        str: matched filenames in a directory
    """
    path = pathlib.Path(data_root)
    for fp in path.glob(pattern):
        yield fp


def load_table(json_file, table_name, version, **kwargs):
    """Load records from json files into a pandas table

    Args:
        json_file (str): filename to json.
        table_name (str): table name in the json file to be loaded
        version (str): requested version of this table
        **kwargs: arbitrary keyword arguments to be passed to pandas'
            json_normalize method.

    Returns:
        a pandas dataframe of the loaded table.

    Raises:
        VersionError: If the version in json file does not match the requested
        version.
    """
    logger.debug(f"Loading table {table_name} from {json_file}")
    data = json.load(open(json_file, "r"))
    verify_version(data, version)
    table = pd.json_normalize(data[table_name], **kwargs)

    return table



import math

import numpy as np
from pyquaternion import Quaternion


def group_bbox2d_per_label(bboxes):
    """Group 2D bounding boxes with same label.

    Args:
        bboxes (list[BBox2D]): a list of 2D bounding boxes

    Returns:
        dict: a dictionary of 2d boundign box group.
        {label1: [bbox1, bboxes2, ...], label2: [bbox1, ...]}
    """
    bboxes_per_label = {}
    for box in bboxes:
        if box.label not in bboxes_per_label:
            bboxes_per_label[box.label] = []
        bboxes_per_label[box.label].append(box)

    return bboxes_per_label


class BBox2D:
    """Canonical Representation of a 2D bounding box.

    Attributes:
        label (str): string representation of the label.
        x (float): x pixel coordinate of the upper left corner.
        y (float): y pixel coordinate of the upper left corner.
        w (float): width (number of pixels)of the bounding box.
        h (float): height (number of pixels) of the bounding box.
        score (float): detection confidence score. Default is set to score=1.
            if this is a ground truth bounding box.

    Examples:
        Here is an example about how to use this class.

        .. code-block::

            >>> gt_bbox = BBox2D(label='car', x=2, y=6, w=2, h=4)
            >>> gt_bbox
            "label='car'|score=1.0|x=2.0|y=6.0|w=2.0|h=4.0"
            >>> pred_bbox = BBox2D(label='car', x=2, y=5, w=2, h=4, score=0.79)
            >>> pred_bbox.area
            8
            >>> pred_bbox.intersect_with(gt_bbox)
            True
            >>> pred_bbox.intersection(gt_bbox)
            6
            >>> pred_bbox.union(gt_bbox)
            10
            >>> pred_bbox.iou(gt_bbox)
            0.6

    """

    def __init__(self, label, x, y, w, h, score=1.0):
        """ Initialize 2D bounding box object

        Args:
            label (str): string representation of the label
            x (float): x pixel coordinate of the upper left corner
            y (float): y pixel coordinate of the upper left corner
            w (float): width (number of pixels)of the bounding box
            h (float): height (number of pixels) of the bounding box
            score (float): detection confidence score
        """
        self.label = label
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.score = score

    def __repr__(self):
        return (
            f"label={self.label}|score={self.score:.2f}|"
            f"x={self.x:.2f}|y={self.y:.2f}|w={self.w:.2f}|h={self.h:.2f}"
        )

    def __eq__(self, other):
        return (
            self.x == other.x
            and self.y == other.y
            and self.w == other.w
            and self.h == other.h
            and self.label == other.label
            and math.isclose(self.score, other.score, rel_tol=1e-07)
        )

    @property
    def area(self):
        """Calculate area of this bounding box

        Returns:
            width x height of the bound box
        """
        return self.w * self.h

    def intersect_with(self, other):
        """Check whether this box intersects with other bounding box

        Args:
            other (BBox2D): other bounding box object to check intersection

        Returns:
            True if two bounding boxes intersect, False otherwise
        """
        if self.x > other.x + other.w:
            return False
        if other.x > self.x + self.w:
            return False
        if self.y + self.h < other.y:
            return False
        if self.y > other.y + other.h:
            return False
        return True

    def intersection(self, other):
        """Calculate the intersection area with other bounding box

        Args:
            other (BBox2D): other bounding box object to calculate intersection

        Returns:
            float of the intersection area for two bounding boxes
        """
        x1 = max(self.x, other.x)
        y1 = max(self.y, other.y)
        x2 = min(self.x + self.w, other.x + other.w)
        y2 = min(self.y + self.h, other.y + other.h)
        return (x2 - x1) * (y2 - y1)

    def union(self, other, intersection_area=None):
        """Calculate union area with other bounding box

        Args:
            other (BBox2D): other bounding box object to calculate union
            intersection_area (float): pre-calculated area of intersection

        Returns:
            float of the union area for two bounding boxes
        """
        area_a = self.area
        area_b = other.area
        if not intersection_area:
            intersection_area = self.intersection(other)
        return float(area_a + area_b - intersection_area)

    def iou(self, other):
        """Calculate intersection over union area with other bounding box

        .. math::
                IOU = \\frac{intersection}{union}

        Args:
            other (BBox2D): other bounding box object to calculate iou

        Returns:
            float of the union area for two bounding boxes
        """
        # if boxes don't intersect
        if not self.intersect_with(other):
            return 0
        intersection_area = self.intersection(other)
        union_area = self.union(other, intersection_area=intersection_area)
        # intersection over union
        iou = intersection_area / union_area
        return iou


class BBox3D:
    """
    Class for 3d bounding boxes which can either be predictions or
    ground-truths. This class is the primary representation in this repo of 3d
    bounding boxes and is based off of the Nuscenes style dataset.
    """

    def __init__(
        self,
        translation,
        size,
        label,
        sample_token,
        score=1,
        rotation: Quaternion = Quaternion(),
        velocity=(np.nan, np.nan, np.nan),
    ):
        self.sample_token = sample_token
        self.translation = translation
        self.size = size
        self.width, self.height, self.length = size
        self.rotation = rotation
        self.velocity = velocity
        self.label = label
        self.score = score

    def _local2world_coordinate(self, x):
        """

        Args:
            x: vector describing point (x,y,z) in local coordinates (where the
            center of the box is 0,0,0)

        Returns: the x,y,z coordinates of the input point in global coordinates

        """

        y = np.array(self.translation) + self.rotation.rotate(x)
        return y

    @property
    def back_left_bottom_pt(self):
        """

        Returns: :py:class:`float`: Back-left-bottom point.

        """
        p = np.array([-self.width / 2, -self.height / 2, -self.length / 2])
        p = self._local2world_coordinate(p)
        return p

    @property
    def front_left_bottom_pt(self):
        """
        :py:class:`float`: Front-left-bottom point.
        """
        p = np.array([-self.width / 2, -self.height / 2, self.length / 2])
        p = self._local2world_coordinate(p)
        return p

    @property
    def front_right_bottom_pt(self):
        """
        :py:class:`float`: Front-right-bottom point.
        """
        p = np.array([self.width / 2, -self.height / 2, self.length / 2])
        p = self._local2world_coordinate(p)
        return p

    @property
    def back_right_bottom_pt(self):
        """
        :py:class:`float`: Back-right-bottom point.
        """
        p = np.array([self.width / 2, -self.height / 2, -self.length / 2])
        p = self._local2world_coordinate(p)
        return p

    @property
    def back_left_top_pt(self):
        """
        :py:class:`float`: Back-left-top point.
        """
        p = np.array([-self.width / 2, self.height / 2, -self.length / 2])
        p = self._local2world_coordinate(p)
        return p

    @property
    def front_left_top_pt(self):
        """
        :py:class:`float`: Front-left-top point.
        """
        p = np.array([-self.width / 2, self.height / 2, self.length / 2])
        p = self._local2world_coordinate(p)
        return p

    @property
    def front_right_top_pt(self):
        """
        :py:class:`float`: Front-right-top point.
        """
        p = np.array([self.width / 2, self.height / 2, self.length / 2])
        p = self._local2world_coordinate(p)
        return p

    @property
    def back_right_top_pt(self):
        """
        :py:class:`float`: Back-right-top point.
        """
        p = np.array([self.width / 2, self.height / 2, -self.length / 2])
        p = self._local2world_coordinate(p)
        return p

    @property
    def p(self) -> np.ndarray:
        """

        Returns: list of all 8 corners of the box beginning with the the bottom
         four corners and then the top
        four corners, both in counterclockwise order (from birds eye view)
        beginning with the back-left corner

        """
        x = np.vstack(
            [
                self.back_left_bottom_pt,
                self.front_left_bottom_pt,
                self.front_right_bottom_pt,
                self.back_right_bottom_pt,
                self.back_left_top_pt,
                self.front_left_top_pt,
                self.front_right_top_pt,
                self.back_right_top_pt,
            ]
        )
        return x



""" Load Synthetic dataset references tables
"""
import pandas as pd

from .tables import DATASET_TABLES, SCHEMA_VERSION, glob, load_table
from .validation import NoRecordError


class AnnotationDefinitions:
    """Load annotation_definitions table

    For more detail, see schema design here:
    :ref:`annotation_definitions.json`

    Attributes:
        table (pd): a collection of annotation_definitions records
    """

    TABLE_NAME = "annotation_definitions"
    FILE_PATTERN = DATASET_TABLES[TABLE_NAME].file

    def __init__(self, data_root, version=SCHEMA_VERSION):
        """ Initialize AnnotationDefinitions

        Args:
            data_root (str): the root directory of the dataset containing
        tables
            version (str): desired schema version
        """
        self.table = self.load_annotation_definitions(data_root, version)

    def load_annotation_definitions(self, data_root, version):
        """Load annotation definition files.

        For more detail, see schema design here:
        :ref:`annotation_definitions.json`

        Args:
            data_root (str): the root directory of the dataset containing
        tables
            version (str): desired schema version

        Returns:
            A Pandas dataframe with annotation definition records.
            Columns: 'id' (annotation id), 'name' (annotation name),
            'description' (string description), 'format'
            (string describing format), 'spec' ( Format-specific specification
            for the annotation values)
        """
        definitions = []
        for def_file in glob(data_root, self.FILE_PATTERN):
            definition = load_table(def_file, self.TABLE_NAME, version)
            definitions.append(definition)

        if definitions:
            combined = pd.concat(definitions, axis=0).drop_duplicates(
                subset="id"
            )
        else:
            combined = pd.DataFrame({})

        return combined

    def get_definition(self, def_id):
        """Get the annotation definition for a given definition id

        Args:
            def_id (int): annotation definition id used to filter results

        Returns:
            a dictionary containing the annotation definition
        """
        mask = self.table.id == def_id
        definition = self.table[mask]
        if definition.empty:
            raise NoRecordError(
                f"No records are found in the annotation_definitions file "
                f"that matches the specified definition id: {def_id}"
            )
        definition = definition.to_dict("records")[0]

        return definition


class MetricDefinitions:
    """Load metric_definitions table

    For more detail, see schema design here:

    :ref:`metric_definitions.json`

    Attributes:
        table (pd): a collection of metric_definitions records with columns: id
    (id for metric definition), name, description, spec (definition specific
    spec)
    """

    TABLE_NAME = "metric_definitions"
    FILE_PATTERN = DATASET_TABLES[TABLE_NAME].file

    def __init__(self, data_root, version=SCHEMA_VERSION):
        """ Initialize MetricDefinitions
        Args:
            data_root (str): the root directory of the dataset containing
        tables
            version (str): desired schema version
        """
        self.table = self.load_metric_definitions(data_root, version)

    def load_metric_definitions(self, data_root, version):
        """Load metric definition files.

        :ref:`metric_definitions.json`

        Args:
            data_root (str): the root directory of the dataset containing tables
            version (str): desired schema version

        Returns:
            A Pandas dataframe with metric definition records.
        a collection of metric_definitions records with columns: id
    (id for metric definition), name, description, spec (definition specific
    spec)
        """
        definitions = []
        for def_file in glob(data_root, self.FILE_PATTERN):
            definition = load_table(def_file, self.TABLE_NAME, version)
            definitions.append(definition)

        combined = pd.concat(definitions, axis=0).drop_duplicates(subset="id")

        return combined

    def get_definition(self, def_id):
        """Get the metric definition for a given definition id

        Args:
            def_id (int): metric definition id used to filter results

        Returns:
            a dictionary containing metric definition
        """
        mask = self.table.id == def_id
        definition = self.table[mask]
        if definition.empty:
            raise NoRecordError(
                f"No records are found in the metric_definitions file "
                f"that matches the specified definition id: {def_id}"
            )
        definition = definition.to_dict("records")[0]

        return definition


class Egos:
    """Load egos table

    For more detail, see schema design here:
    :ref:`egos.json`

    Attributes:
        table (pd): a collection of egos records
    """

    TABLE_NAME = "egos"
    FILE_PATTERN = DATASET_TABLES[TABLE_NAME].file

    def __init__(self, data_root, version=SCHEMA_VERSION):
        """Initialize `:ref:Egos`


        Args:
            data_root (str): the root directory of the dataset containing
            ego tables. Two columns: id (ego id) and description
            version (str): desired schema version
        """
        self.table = self.load_egos(data_root, version)

    def load_egos(self, data_root, version):
        """Load egos files.
        For more detail, see schema design here:

        :ref:`egos.json`

        Args:
            data_root (str): the root directory of the dataset containing
            ego tables
            version (str): desired schema version

        Returns:
            A pandas dataframe with all ego records with two columns: id
            (ego id) and description
        """
        egos = []
        for ego_file in glob(data_root, self.FILE_PATTERN):
            ego = load_table(ego_file, self.TABLE_NAME, version)
            egos.append(ego)
        combined = pd.concat(egos, axis=0).drop_duplicates(subset="id")

        return combined


class Sensors:
    """Load sensors table

    For more detail, see schema design here:

    :ref:`sensors.json`

    Attributes:
        table (pd): a collection of sensors records with columns:
        'id' (sensor id), 'ego_id', 'modality'
        ({camera, lidar, radar, sonar,...} -- Sensor modality), 'description'

    """

    TABLE_NAME = "sensors"
    FILE_PATTERN = DATASET_TABLES[TABLE_NAME].file

    def __init__(self, data_root, version=SCHEMA_VERSION):
        """ Initialize Sensors

        Args:
            data_root (str): the root directory of the dataset containing
        tables
            version (str): desired schema version
        """
        self.table = self.load_sensors(data_root, version)

    def load_sensors(self, data_root, version):
        """Load sensors files.

        For more detail, see schema design here:

        :ref:`sensors.json`

        Args:
            data_root (str): the root directory of the dataset containing
        tables
            version (str): desired schema version

        Returns:
            A pandas dataframe with all sensors records  with columns:
        'id' (sensor id), 'ego_id', 'modality'
        ({camera, lidar, radar, sonar,...} -- Sensor modality), 'description'
        """
        sensors = []
        for sensor_file in glob(data_root, self.FILE_PATTERN):
            sensor = load_table(sensor_file, self.TABLE_NAME, version)
            sensors.append(sensor)
        combined = pd.concat(sensors, axis=0).drop_duplicates(subset="id")

        return combined



import argparse
import json
import os

import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output

import datasetinsights.stats.visualization.overview as overview
from datasetinsights.stats.visualization.app import get_app
from datasetinsights.stats.visualization.object_detection import (
    render_object_detection_layout,
)

app = get_app()


def main_layout():
    """ Method for generating main app layout.

    Returns:
        html layout: main layout design with tabs for overview statistics
            and object detection.
    """
    app_layout = html.Div(
        [
            html.H1(
                children="Dataset Insights",
                style={
                    "textAlign": "center",
                    "padding": 20,
                    "background": "lightgrey",
                },
            ),
            html.Div(
                [
                    dcc.Tabs(
                        id="page_tabs",
                        value="dataset_overview",
                        children=[
                            dcc.Tab(
                                label="Overview", value="dataset_overview",
                            ),
                            dcc.Tab(
                                label="Object Detection",
                                value="object_detection",
                            ),
                        ],
                    ),
                    html.Div(id="main_page_tabs"),
                ]
            ),
            # Sharing data between callbacks using hidden division.
            # These hidden dcc and html components are for storing data-root
            # into the division. This is further used in callbacks made in the
            # object_detection module. This is a temporary hack and can be found
            # in example 1 of sharing data between callback dash tutorial.
            # ref: https://dash.plotly.com/sharing-data-between-callbacks
            # TODO: Fix this using a better solution to share data.
            dcc.Dropdown(id="dropdown", style={"display": "none"}),
            html.Div(id="data_root_value", style={"display": "none"}),
        ]
    )
    return app_layout


@app.callback(
    Output("data_root_value", "children"), [Input("dropdown", "value")]
)
def store_data_root(value):
    """ Method for storing data-root value in a hidden division.

    Returns:
        json : data-root encoded in json to be stored in data_root_value div.
    """
    json_data_root = json.dumps(data_root)

    return json_data_root


@app.callback(
    Output("main_page_tabs", "children"),
    [Input("page_tabs", "value"), Input("data_root_value", "children")],
)
def render_content(value, json_data_root):
    """ Method for rendering dashboard layout based
        on the selected tab value.

    Args:
        value(str): selected tab value
        json_data_root: data root stored in hidden div in json format.

    Returns:
        html layout: layout for the selected tab.
    """
    # read data root value from the data_root_value division
    data_root = json.loads(json_data_root)
    if value == "dataset_overview":
        return overview.html_overview(data_root)
    elif value == "object_detection":
        return render_object_detection_layout(data_root)


def check_path(path):
    """ Method for checking if the given data-root path is valid or not."""
    if os.path.isdir(path):
        return path
    else:
        raise ValueError(f"Path {path} not found")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--data-root", help="Path to the data root")
    args = parser.parse_args()
    data_root = check_path(args.data_root)
    app.layout = main_layout()
    app.run_server(debug=True)



""" Validate Simulation Data
"""


class VersionError(Exception):
    """Raise when the data file version does not match"""

    pass


class DuplicateRecordError(Exception):
    """ Raise when the definition file has duplicate definition id
    """

    pass


class NoRecordError(Exception):
    """ Raise when no record is found matching a given definition id
    """

    pass


def verify_version(json_data, version):
    """Verify json schema version

    Args:
        json_data (json): a json object loaded from file.
        version (str): string of the requested version.

    Raises:
        VersionError: If the version in json file does not match the requested
    version.
    """
    loaded = json_data["version"]
    if loaded != version:
        raise VersionError(f"Version mismatch. Expected version: {version}")


def check_duplicate_records(table, column, table_name):
    """ Check if table has duplicate records for a given column

    Args:
        table (pd.DataFrame): a pandas dataframe
        column (str): the column where no duplication is allowed
        table_name (str): table name

    Raises:
        DuplicateRecordError: If duplicate records are found in a column
    """
    if table[column].nunique() != len(table):
        raise DuplicateRecordError(
            f"Duplicate record was found in {column} of table {table_name}. "
            f"This column is expected to be unique. Violating this requirement "
            f"might cause ambiguity when the records are loaded."
        )



import dash_core_components as dcc
import dash_html_components as html
import dash_table
import pandas as pd

import datasetinsights.datasets.unity_perception as sim
import datasetinsights.stats.visualization.constants as constants

from .plots import histogram_plot, rotation_plot


class ScaleFactor:
    """Generate scale factor distribution.

    Scale Factor describes the size of the rendered object in a capture.
    Higher the scale factor, higher would be the visible pixels.

    Atrributes:
        captures(sim.Captures): a collection of capture records.

    """

    def __init__(self, data_root):
        """

        Args:
            data_root(str): path to the dataset.

        """
        capture = sim.Captures(data_root)
        self.captures = capture.filter(
            def_id=constants.BOUNDING_BOX_2D_DEFINITION_ID
        )

    @staticmethod
    def generate_scale_data(captures):
        """ Method to extract scale parameter from sensor data.

        Args:
            captures(sim.Captures): a collection of capture records.

        Returns:
            pandas.DataFrame: contains 'scale' parameter from the sensor data.

        """
        df_scale_factor = pd.DataFrame(
            [sensor_data["scale"] for sensor_data in captures["sensor"]],
            columns=["scale"],
        )
        return df_scale_factor

    def _generate_scale_factor_figures(self):
        """ Method for generating plots for scale factor distribution.

        Returns:
            plotly.graph_objects.Figure: scale factor distribution.

        """
        df_scale_factor = self.generate_scale_data(self.captures)
        scale_factor_distribution_figure = histogram_plot(
            df_scale_factor,
            x="scale",
            x_title="Scale",
            y_title="Capture count",
            title="Distribution of Scale Factor",
            range_x=[
                min(df_scale_factor["scale"]),
                max(df_scale_factor["scale"]),
            ],
            max_samples=constants.MAX_SAMPLES,
        )
        return scale_factor_distribution_figure

    def html(self):
        """ Method for generating plots for scale factor distribution.

        Returns:
            html layout: displays scale factor distribution.

        """
        scale_factor_distribution_figure = self._generate_scale_factor_figures()
        html_layout = html.Div(
            [
                dcc.Markdown(
                    """# Scale Factor """, style={"text-align": "center"}
                ),
                dcc.Graph(
                    id="scale_factor_distribution_figure",
                    figure=scale_factor_distribution_figure,
                ),
            ]
        )
        return html_layout


class UserParameter:
    """Generate User Parameter

    Generate User Parameter table to be displayed on the Dashboard.
    Users parameters, such as ScaleFactors, MaxFrames,
    MaxForegroundObjectsPerFrame are used to control the domain
    randomization parameter used in the simulation.

    Atrributes:
        metrics(sim.Metrics): a collection of metrics records
        user_parameter_table (pandas.DataFrame): dataframe containing user
            parameters.

    Args:
        data_root(str): path to the dataset.
    """

    def __init__(self, data_root):
        self.metrics = sim.Metrics(data_root=data_root)
        self.user_parameter_table = self.metrics.filter_metrics(
            constants.USER_PARAMETERS_DEFINITION_ID
        )

    def html(self):
        """ Method for generating html layout for the
            user input parameter table.

        Returns:
            html layout: displays user input parameter table.

        """
        html_layout = html.Div(
            [
                dcc.Markdown(
                    """# User Input Parameters Table """,
                    style={"text-align": "center"},
                ),
                dash_table.DataTable(
                    id="user_parameter_table",
                    columns=[
                        {"name": i, "id": i}
                        for i in self.user_parameter_table.columns
                    ],
                    data=self.user_parameter_table.to_dict("records"),
                    style_table={"overflowX": "auto"},
                    style_cell={
                        "height": "auto",
                        # all three widths are needed
                        "minWidth": "180px",
                        "width": "180px",
                        "maxWidth": "400px",
                        "whiteSpace": "normal",
                        "text-align": "center",
                    },
                ),
            ]
        )
        return html_layout


class Lighting:
    """ This class contains methods for object lighting statistics
    visualization.

    Attributes:
        metrics(sim.Metrics): a collection of metrics records
        lighting (pandas.DataFrame): contains information about per-frame light
            color and orientation information.

    """

    X_Y_COLUMNS = ["x_rotation", "y_rotation"]
    COLOR_COLUMNS = ["color.r", "color.g", "color.b", "color.a"]

    def __init__(self, data_root):
        """
        Args:
            data_root(str): path to the dataset.

        """
        self.metrics = sim.Metrics(data_root=data_root)
        self.lighting = self._read_lighting_info()

    def _read_lighting_info(self):
        """ Method to obtain per-frame light color and orientation information.

        Returns:
            pandas.DataFrame: contains information about per-frame light color
                and orientation information.

        """
        metrics = self.metrics
        filtered_metrics = metrics.filter_metrics(
            constants.LIGHTING_INFO_DEFINITION_ID
        )
        colors = pd.json_normalize(filtered_metrics["color"])
        colors.columns = self.COLOR_COLUMNS
        combined = pd.concat(
            [filtered_metrics[self.X_Y_COLUMNS], colors], axis=1, join="inner"
        )

        return combined

    def _generate_figures_lighting(self):
        """ Method for generating plots for displaying light orientations,
            rotations and color distribution.

        Args:
            lighting(pandas.DataFrame): contains lighting information.

        Returns:
            plotly.graph_objects.Figure: chart to display light orientations,
                rotations and color distribution.

        """

        lighting_fig = rotation_plot(
            self.lighting,
            x="x_rotation",
            y="y_rotation",
            max_samples=constants.MAX_SAMPLES,
        )

        lighting_x_rot_fig = histogram_plot(
            self.lighting,
            x="x_rotation",
            x_title="Lighting Rotation (Degree)",
            y_title="Frequency",
            title="Distribution of Lighting Rotations along X direction",
            max_samples=constants.MAX_SAMPLES,
        )

        lighting_y_rot_fig = histogram_plot(
            self.lighting,
            x="y_rotation",
            x_title="Lighting Rotation (Degree)",
            y_title="Frequency",
            title="Distribution of Lighting Rotations along Y direction",
            max_samples=constants.MAX_SAMPLES,
        )

        lighting_redness_fig = histogram_plot(
            self.lighting,
            x="color.r",
            x_title="Lighting Color",
            y_title="Frequency",
            title="Distribution of Lighting Color Redness",
            max_samples=constants.MAX_SAMPLES,
            color_discrete_sequence=["indianred"],
        )

        lighting_greeness_fig = histogram_plot(
            self.lighting,
            x="color.g",
            x_title="Lighting Color",
            y_title="Frequency",
            title="Distribution of Lighting Color Greeness",
            max_samples=constants.MAX_SAMPLES,
            color_discrete_sequence=["MediumSeaGreen"],
        )

        lighting_blueness_fig = histogram_plot(
            self.lighting,
            x="color.b",
            x_title="Lighting Color",
            y_title="Frequency",
            title="Distribution of Lighting Color Blueness",
            max_samples=constants.MAX_SAMPLES,
        )

        return {
            "lighting_fig": lighting_fig,
            "lighting_x_rot_fig": lighting_x_rot_fig,
            "lighting_y_rot_fig": lighting_y_rot_fig,
            "lighting_redness_fig": lighting_redness_fig,
            "lighting_greeness_fig": lighting_greeness_fig,
            "lighting_blueness_fig": lighting_blueness_fig,
        }

    def html(self):
        """ Method for generating html layout for the
            lighting statistics.

        Returns:
            html layout: displays lighting graphs.

        """
        lighting_figures = self._generate_figures_lighting()

        html_layout = html.Div(
            [
                html.Div(
                    [
                        dcc.Markdown(
                            """# Lighting Orientations """,
                            style={"text-align": "center"},
                        ),
                        dcc.Graph(
                            id="lighting_fig",
                            figure=lighting_figures["lighting_fig"],
                        ),
                        dcc.Tabs(
                            [
                                dcc.Tab(
                                    label="Lighting along X direction",
                                    children=[
                                        dcc.Graph(
                                            id="lighting_x_rot_fig",
                                            figure=lighting_figures[
                                                "lighting_x_rot_fig"
                                            ],
                                        )
                                    ],
                                ),
                                dcc.Tab(
                                    label="Lighting along Y direction",
                                    children=[
                                        dcc.Graph(
                                            id="lighting_y_rot_fig",
                                            figure=lighting_figures[
                                                "lighting_y_rot_fig"
                                            ],
                                        )
                                    ],
                                ),
                            ]
                        ),
                    ]
                ),
                html.Div(
                    [
                        dcc.Graph(
                            id="lighting_redness_fig",
                            figure=lighting_figures["lighting_redness_fig"],
                        ),
                        dcc.Graph(
                            id="lighting_greeness_fig",
                            figure=lighting_figures["lighting_greeness_fig"],
                        ),
                        dcc.Graph(
                            id="lighting_blueness_fig",
                            figure=lighting_figures["lighting_blueness_fig"],
                        ),
                    ],
                    style={"columnCount": 3},
                ),
            ]
        )
        return html_layout


class ObjectPlacement:
    """ This class contains methods for object orientation statistics
    visualization.

    Attributes:
        metrics(sim.Metrics): a collection of metrics records
        lighting (pandas.DataFrame): contains information about per-frame light
            color and orientation information.

    """

    OBJECT_ORIENTATION = ("x_rot", "y_rot", "z_rot")

    def __init__(self, data_root):
        """

        Args:
            data_root(str): path to the dataset.

        """
        self.metrics = sim.Metrics(data_root=data_root)
        self.rotation = self._read_foreground_placement_info()

    def _read_foreground_placement_info(self):
        """ Method to obtain rotations of the foreground objects.

        Returns:
            pandas.DataFrame: contains information about foreground object
                orientations.

        """
        metrics = self.metrics
        filtered_metrics = metrics.filter_metrics(
            constants.FOREGROUND_PLACEMENT_INFO_DEFINITION_ID
        )
        combined = pd.DataFrame(
            filtered_metrics["rotation"].to_list(),
            columns=self.OBJECT_ORIENTATION,
        )

        return combined

    def _generate_figures_orientation(self, orientation):
        """ Method for generating object orientation statistics

        Args:
            orientation(pandas.DataFrame): contains object orientations
                along X, Y, Z direction.

        Returns:
            plotly.graph_objects.Figure: charts for object orientation
                statistics.

        """
        orientation_rotation_plot_fig = rotation_plot(
            orientation,
            x="x_rot",
            y="y_rot",
            z="z_rot",
            max_samples=constants.MAX_SAMPLES,
        )

        rotation_x_dir_fig = histogram_plot(
            orientation,
            x="x_rot",
            x_title="Object Rotation (Degree)",
            y_title="Frequency",
            title="Distribution of Object Rotations along X direction",
            max_samples=constants.MAX_SAMPLES,
        )

        rotation_y_dir_fig = histogram_plot(
            orientation,
            x="y_rot",
            x_title="Object Rotation (Degree)",
            y_title="Frequency",
            title="Distribution of Object Rotations along Y direction",
            max_samples=constants.MAX_SAMPLES,
        )

        rotation_z_dir_fig = histogram_plot(
            orientation,
            x="z_rot",
            x_title="Object Rotation (Degree)",
            y_title="Frequency",
            title="Distribution of Object Rotations along Z direction",
            max_samples=constants.MAX_SAMPLES,
        )

        return {
            "orientation_rotation_plot_fig": orientation_rotation_plot_fig,
            "rotation_x_dir_fig": rotation_x_dir_fig,
            "rotation_y_dir_fig": rotation_y_dir_fig,
            "rotation_z_dir_fig": rotation_z_dir_fig,
        }

    def html(self):
        """ Method for generating html layout for the object
            orientation statistics.

        Returns:
            html layout: displays object orientation graphs.

        """
        orientation_figures = self._generate_figures_orientation(self.rotation)

        html_layout = html.Div(
            [
                dcc.Markdown(
                    """# Object Orientations """, style={"text-align": "center"}
                ),
                dcc.Graph(
                    id="orientation_rotation",
                    figure=orientation_figures["orientation_rotation_plot_fig"],
                ),
                dcc.Tabs(
                    [
                        dcc.Tab(
                            label="Rotation along X direction",
                            children=[
                                dcc.Graph(
                                    id="rotation_x_dir",
                                    figure=orientation_figures[
                                        "rotation_x_dir_fig"
                                    ],
                                )
                            ],
                        ),
                        dcc.Tab(
                            label="Rotation along Y direction",
                            children=[
                                dcc.Graph(
                                    id="rotation_y_dir",
                                    figure=orientation_figures[
                                        "rotation_y_dir_fig"
                                    ],
                                )
                            ],
                        ),
                        dcc.Tab(
                            label="Rotation along Z direction",
                            children=[
                                dcc.Graph(
                                    id="rotation_z_dir",
                                    figure=orientation_figures[
                                        "rotation_z_dir_fig"
                                    ],
                                )
                            ],
                        ),
                    ]
                ),
            ],
        )
        return html_layout


def render_object_detection_layout(data_root):
    """ Method for displaying object detection statistics.

    Args:
        data_root(str): path to the dataset.

    Returns:
        html layout: displays graphs for rotation and
            lighting statistics for the object.

    """
    user_parameter = UserParameter(data_root)
    object_placement = ObjectPlacement(data_root)
    lighting = Lighting(data_root)
    scale_factor = ScaleFactor(data_root)
    object_detection_layout = html.Div(
        [
            html.Div(id="object_detection"),
            user_parameter.html(),
            object_placement.html(),
            lighting.html(),
            scale_factor.html(),
        ]
    )
    return object_detection_layout



import logging

import click

from datasetinsights.commands import Entrypoint
from datasetinsights.constants import CONTEXT_SETTINGS

logging.basicConfig(
    level=logging.INFO,
    format=(
        "%(levelname)s | %(asctime)s | %(name)s | %(threadName)s | "
        "%(message)s"
    ),
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger(__name__)


@click.command(
    cls=Entrypoint, help="Dataset Insights.", context_settings=CONTEXT_SETTINGS,
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Enables verbose mode.",
)
def entrypoint(verbose):
    if verbose:
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)


if __name__ == "__main__":
    entrypoint()



class DownloadError(Exception):
    """ Raise when download file failed.
    """


class ChecksumError(Exception):
    """ Raises when the downloaded file checksum is not correct.
    """


class InvalidTrackerError(Exception):
    """ Raises when unknown tracker requested .
    """



import os

import dash


def _init_app():
    """ Intializes the dash app."""

    this_dir = os.path.dirname(os.path.abspath(__file__))
    css_file = os.path.join(this_dir, "stylesheet.css")
    app = dash.Dash(
        __name__,
        external_stylesheets=[css_file],
        suppress_callback_exceptions=True,
    )
    return app


_app = _init_app()


def get_app():
    return _app



import os

import click


class Entrypoint(click.MultiCommand):
    """ Click MultiCommand Entrypoint For Datasetinsights CLI
    """

    def list_commands(self, ctx):
        """Dynamically get the list of commands."""
        rv = []
        for filename in os.listdir(os.path.dirname(__file__)):
            if filename.endswith(".py") and not filename.startswith("__init__"):
                rv.append(filename[:-3])
        rv.sort()

        return rv

    def get_command(self, ctx, name):
        """Dynamically get the command."""
        ns = {}
        fn = os.path.join(os.path.dirname(__file__), name + ".py")
        if not os.path.exists(fn):
            return None
        with open(fn) as f:
            code = compile(f.read(), fn, "exec")
            eval(code, ns, ns)

        return ns["cli"]



from .captures import Captures
from .metrics import Metrics
from .references import AnnotationDefinitions, Egos, MetricDefinitions, Sensors

__all__ = [
    "AnnotationDefinitions",
    "Captures",
    "Egos",
    "Metrics",
    "MetricDefinitions",
    "Sensors",
]



import re
from abc import ABC, abstractmethod

_registry = {}


def _find_downloader(source_uri):
    """
     This function returns the correct DatasetDownloader
     from a registry based on the source-uri provided

        Args:
            source_uri: URI of where this data should be downloaded.

        Returns: The dataset downloader class that is registered with the
                 source-uri protocol.

    """
    protocols = "|".join(_registry.keys())
    pattern = re.compile(f"({protocols})")

    protocol = pattern.findall(source_uri)

    if source_uri.startswith(("https://", "http://")):
        protocol = "http://"
    elif protocol:
        protocol = protocol[0]
    else:
        raise ValueError(f"Downloader not found for source-uri '{source_uri}'")

    return _registry.get(protocol)


def create_dataset_downloader(source_uri, **kwargs):
    """
    This function instantiates the dataset downloader
     after finding it with the source-uri provided

    Args:
        source_uri: URI used to look up the correct dataset downloader
        **kwargs:

    Returns: The dataset downloader instance matching the source-uri.

    """
    downloader_class = _find_downloader(source_uri=source_uri)
    return downloader_class(**kwargs)


class DatasetDownloader(ABC):
    """This is the base class for all dataset downloaders
    The DatasetDownloader can be subclasses in the following way

    class NewDatasetDownloader(DatasetDownloader, protocol="protocol://")

    Here the 'protocol://' should match the prefix that the method download
    source_uri supports. Example http:// gs://

    """

    def __init__(self, **kwargs):
        pass

    @classmethod
    def __init_subclass__(cls, protocol=None, **kwargs):
        if protocol:
            _registry[protocol] = cls
        else:
            raise NotImplementedError(
                f"Subclass needs to have class keyword argument named protocol."
            )
        super().__init_subclass__(**kwargs)

    @abstractmethod
    def download(self, source_uri, output, **kwargs):
        """ This method downloads a dataset stored at the source_uri and stores it
        in the output directory

        Args:
            source_uri: URI that points to the dataset that should be downloaded
            output: path to local folder where the dataset should be stored
        """
        raise NotImplementedError("Subclass needs to implement this method")



""" Use a bounding box library to plot pretty bounding boxes
with a simple Python API. This library helps to display pretty bounding boxes
with a chosen set of colors.
Reference: https://github.com/nalepae/bounding-box
"""
import os as _os
import pathlib
import random
from hashlib import md5 as _md5

import cv2 as _cv2
import numpy as _np
from PIL import ImageFont

FONT_PATH = _os.path.join(
    pathlib.Path(__file__).parent.absolute(), "font", "arial.ttf"
)
_COLOR_NAME_TO_RGB = dict(
    navy=((0, 38, 63), (119, 193, 250)),
    blue=((0, 120, 210), (173, 220, 252)),
    aqua=((115, 221, 252), (0, 76, 100)),
    teal=((15, 205, 202), (0, 0, 0)),
    olive=((52, 153, 114), (25, 58, 45)),
    green=((0, 204, 84), (15, 64, 31)),
    lime=((1, 255, 127), (0, 102, 53)),
    yellow=((255, 216, 70), (103, 87, 28)),
    orange=((255, 125, 57), (104, 48, 19)),
    red=((255, 47, 65), (131, 0, 17)),
    maroon=((135, 13, 75), (239, 117, 173)),
    fuchsia=((246, 0, 184), (103, 0, 78)),
    purple=((179, 17, 193), (241, 167, 244)),
    gray=((168, 168, 168), (0, 0, 0)),
    silver=((220, 220, 220), (0, 0, 0)),
)
_COLOR_NAMES = list(_COLOR_NAME_TO_RGB)
_DEFAULT_COLOR_NAME = "green"


def add_single_bbox_on_image(
    image, bbox, label, color, font_size=100, box_line_width=15
):
    """ Add single bounding box with label on a given image.

    Args:
        image (numpy array): a numpy array for an image.
        bbox (BBox2D): a canonical bounding box.
        color (str): a color name for one bounding box.
        If color = None, it will randomly assign a color for each box.
        font_size (int): font size for each label. Defaults to 100.
        box_line_width (int): line width of the bounding boxes. Defaults to 15.
    """
    left, top = (bbox.x, bbox.y)
    right, bottom = (bbox.x + bbox.w, bbox.y + bbox.h)

    _add_single_bbox_on_image(
        image,
        left,
        top,
        right,
        bottom,
        label=label,
        color=color,
        font_size=font_size,
        box_line_width=box_line_width,
    )


def _rgb_to_bgr(color):
    return list(reversed(color))


def _color_image(image, font_color, background_color):
    return background_color + (font_color - background_color) * image / 255


def _get_label_image(
    text, font_color_tuple_bgr, background_color_tuple_bgr, font_size=100
):
    """ Add text and background color for one label.

    Args:
        text (str): label name.
        font_color_tuple_bgr (tuple): font RGB color.
        background_color_tuple_bgr (tuple): background RGB color.
        font_size (int): font size for the label text.

    Returns:
        numpy array: a numpy array for a rendered label.
    """
    _FONT = ImageFont.truetype(FONT_PATH, font_size)
    text_image = _FONT.getmask(text)
    shape = list(reversed(text_image.size))
    bw_image = _np.array(text_image).reshape(shape)

    image = [
        _color_image(bw_image, font_color, background_color)[None, ...]
        for font_color, background_color in zip(
            font_color_tuple_bgr, background_color_tuple_bgr
        )
    ]

    return _np.concatenate(image).transpose(1, 2, 0)


def _add_single_bbox_on_image(
    image,
    left,
    top,
    right,
    bottom,
    label=None,
    color=None,
    font_size=100,
    box_line_width=15,
):
    """ Add single bounding box with label on a given image.

    Add single bounding box and a label text with label on a given image. If the
    label text exceeds the original image border, it would be cropped.
    """
    try:
        left, top, right, bottom = int(left), int(top), int(right), int(bottom)
    except ValueError:
        raise TypeError("'left', 'top', 'right' & 'bottom' must be a number")

    if label and not color:
        hex_digest = _md5(label.encode()).hexdigest()
        color_index = int(hex_digest, 16) % len(_COLOR_NAME_TO_RGB)
        color = _COLOR_NAMES[color_index]
    elif not label:
        color = random.choice(_COLOR_NAMES)

    colors = [list(item) for item in _COLOR_NAME_TO_RGB[color]]
    color, color_text = colors

    _cv2.rectangle(image, (left, top), (right, bottom), color, box_line_width)

    if label:
        label_image = _get_label_image(label, color_text, color, font_size)
        _add_label_on_image(label_image, image, left, top, color)


def _add_label_on_image(label_image, image, left, top, color):
    """Add a label on a bounding box.

    Add a label on a bounding box. Crop the label image if it cross the image
    border.
    """
    image_height, image_width, _ = image.shape
    label_height, label_width, _ = label_image.shape
    rectangle_height, rectangle_width = 1 + label_height, 1 + label_width

    rectangle_bottom = top
    rectangle_left = max(0, min(left - 1, image_width - rectangle_width))

    rectangle_top = rectangle_bottom - rectangle_height
    rectangle_right = rectangle_left + rectangle_width

    label_top = rectangle_top + 1

    if rectangle_top < 0:
        rectangle_top = top
        rectangle_bottom = rectangle_top + label_height + 1

        label_top = rectangle_top

    label_left = rectangle_left + 1
    label_bottom = label_top + label_height
    label_right = label_left + label_width

    rec_left_top = (rectangle_left, rectangle_top)
    rec_right_bottom = (rectangle_right, rectangle_bottom)

    _cv2.rectangle(image, rec_left_top, rec_right_bottom, color, -1)
    _fix_label_at_image_edge(
        label_image, label_left, label_top, label_right, label_bottom, image
    )


def _fix_label_at_image_edge(
    label_image, label_left, label_top, label_right, label_bottom, image
):
    """Fix the label at image edge.

    Crop the label image if it cross the image border.
    """
    image_height, image_width, _ = image.shape
    label_height, label_width, _ = label_image.shape
    label_top = max(0, label_top)
    label_bottom = min(image_height, label_bottom)
    label_left = max(0, label_left)
    label_right = min(image_width, label_right)
    label_actual_width = label_right - label_left
    label_actual_height = label_bottom - label_top
    label_actual_size = label_actual_width * label_actual_height
    if label_actual_size < label_height * label_width:
        image[label_top:label_bottom, label_left:label_right, :] = label_image[
            : (label_bottom - label_top), : (label_right - label_left), :
        ]
    else:
        image[label_top:label_bottom, label_left:label_right, :] = label_image



import base64
import logging
import os
import re
from os import makedirs
from os.path import basename, isdir
from pathlib import Path

from google.cloud.storage import Client

from datasetinsights.io.download import validate_checksum
from datasetinsights.io.exceptions import ChecksumError

logger = logging.getLogger(__name__)


class GCSClient:
    """ This class is used to download data from GCS location
        and perform function such as downloading the dataset and checksum
        validation.
    """

    GCS_PREFIX = "^gs://"
    KEY_SEPARATOR = "/"

    def __init__(self, **kwargs):
        """ Initialize a client to google cloud storage (GCS).
        """
        self.client = Client(**kwargs)

    def download(self, *, url=None, local_path=None, bucket=None, key=None):
        """ This method is used to download the dataset from GCS.

        Args:
            url (str): This is the downloader-uri that indicates where
                              the dataset should be downloaded from.

            local_path (str): This is the path to the directory where the
                          download will store the dataset.

            bucket (str): gcs bucket name
            key (str): object key path

            Examples:
                >>> url = "gs://bucket/folder or gs://bucket/folder/data.zip"
                >>> local_path = "/tmp/folder"
                >>> bucket ="bucket"
                >>> key ="folder/data.zip" or "folder"

        """
        if not (bucket and key) and url:
            bucket, key = self._parse(url)

        bucket_obj = self.client.get_bucket(bucket)
        if self._is_file(bucket_obj, key):
            self._download_file(bucket_obj, key, local_path)
        else:
            self._download_folder(bucket_obj, key, local_path)

    def _download_folder(self, bucket, key, local_path):
        """ download all files from directory
        """
        blobs = bucket.list_blobs(prefix=key)
        for blob in blobs:
            local_file_path = blob.name.replace(key, local_path)
            self._download_validate(blob, local_file_path)

    def _download_file(self, bucket, key, local_path):
        """ download single file
        """
        blob = bucket.get_blob(key)
        key_suffix = key.replace("/" + basename(key), "")
        local_file_path = blob.name.replace(key_suffix, local_path)
        self._download_validate(blob, local_file_path)

    def _download_validate(self, blob, local_file_path):
        """ download file and validate checksum
        """
        self._download_blob(blob, local_file_path)
        self._checksum(blob, local_file_path)

    def _download_blob(self, blob, local_file_path):
        """ download blob from gcs
        Raises:
            NotFound: This will raise when object not found
        """
        dst_dir = local_file_path.replace("/" + basename(local_file_path), "")
        key = blob.name
        if not isdir(dst_dir):
            makedirs(dst_dir)

        logger.info(f"Downloading from {key} to {local_file_path}.")
        blob.download_to_filename(local_file_path)

    def _checksum(self, blob, filename):
        """validate checksum and delete file if checksum does not match

        Raises:
            ChecksumError: This will raise this error if checksum doesn't
                           matches
        """
        expected_checksum = blob.md5_hash
        if expected_checksum:
            expected_checksum_hex = self._md5_hex(expected_checksum)
            try:
                validate_checksum(
                    filename, expected_checksum_hex, algorithm="MD5"
                )
            except ChecksumError as e:
                logger.exception(
                    "Checksum mismatch. Delete the downloaded files."
                )
                os.remove(filename)
                raise e

    def _is_file(self, bucket, key):
        """Check if the key is a file or directory"""
        blob = bucket.get_blob(key)
        return blob and blob.name == key

    def _md5_hex(self, checksum):
        """fix the missing padding if requires and converts into hex"""
        missing_padding = len(checksum) % 4
        if missing_padding != 0:
            checksum += "=" * (4 - missing_padding)
        return base64.b64decode(checksum).hex()

    def _parse(self, url):
        """Split an GCS-prefixed URL into bucket and path."""
        match = re.search(self.GCS_PREFIX, url)
        if not match:
            raise ValueError(
                f"Specified destination prefix: {url} does not start "
                f"with {self.GCS_PREFIX}"
            )
        url = url[len(self.GCS_PREFIX) - 1 :]
        if self.KEY_SEPARATOR not in url:
            raise ValueError(
                f"Specified destination prefix: {self.GCS_PREFIX + url} does "
                f"not have object key "
            )
        idx = url.index(self.KEY_SEPARATOR)
        bucket = url[:idx]
        path = url[(idx + 1) :]

        return bucket, path

    def upload(
        self, *, local_path=None, bucket=None, key=None, url=None, pattern="*"
    ):
        """ Upload a file or list of files from directory to GCS

            Args:
                url (str): This is the gcs location that indicates where
                the dataset should be uploaded.

                local_path (str): This is the path to the directory or file
                where the data is stored.

                bucket (str): gcs bucket name
                key (str): object key path
                pattern: Unix glob patterns. Use **/* for recursive glob.

                Examples:
                    For file upload:
                        >>> url = "gs://bucket/folder/data.zip"
                        >>> local_path = "/tmp/folder/data.zip"
                        >>> bucket ="bucket"
                        >>> key ="folder/data.zip"
                    For directory upload:
                        >>> url = "gs://bucket/folder"
                        >>> local_path = "/tmp/folder"
                        >>> bucket ="bucket"
                        >>> key ="folder"
                        >>> key ="**/*"

        """
        if not (bucket and key) and url:
            bucket, key = self._parse(url)

        bucket_obj = self.client.get_bucket(bucket)
        if isdir(local_path):
            self._upload_folder(
                local_path=local_path,
                bucket=bucket_obj,
                key=key,
                pattern=pattern,
            )
        else:
            self._upload_file(local_path=local_path, bucket=bucket_obj, key=key)

    def _upload_file(self, local_path=None, bucket=None, key=None):
        """ Upload a single object to GCS
        """
        blob = bucket.blob(key)
        logger.info(f"Uploading from {local_path} to {key}.")
        blob.upload_from_filename(local_path)

    def _upload_folder(
        self, local_path=None, bucket=None, key=None, pattern="*"
    ):
        """Upload all files from a folder to GCS based on pattern
        """
        for path in Path(local_path).glob(pattern):
            if path.is_dir():
                continue
            full_path = str(path)
            relative_path = str(path.relative_to(local_path))
            object_key = os.path.join(key, relative_path)
            self._upload_file(
                local_path=full_path, bucket=bucket, key=object_key
            )

    def get_most_recent_blob(self, url=None, bucket_name=None, key=None):
        """ Get the last updated blob in a given bucket under given prefix

            Args:
                bucket_name (str): gcs bucket name
                key (str): object key path
        """
        if not (bucket_name and key) and url:
            bucket_name, key = self._parse(url)

        bucket = self.client.get_bucket(bucket_name)

        if self._is_file(bucket, key):
            # Called on file, return file
            return bucket.get_blob(key)
        else:
            logger.debug(
                f"Cloud path not a file. Checking for most recent file in {url}"
            )
            # Return the blob with the max update time (most recent)
            blobs = self._list_blobs(bucket, prefix=key)
            return max(
                blobs, key=lambda blob: bucket.get_blob(blob.name).updated
            )

    def _list_blobs(self, bucket_name=None, prefix=None):
        """List all blobs with given prefix
        """
        blobs = self.client.list_blobs(bucket_name, prefix=prefix)
        blob_list = list(blobs)
        logger.debug(f"Blobs in {bucket_name} under prefix {prefix}:")
        logger.debug(blob_list)
        return blob_list



""" Load Synthetic dataset captures and annotations tables
"""
import pandas as pd

from datasetinsights.constants import DEFAULT_DATA_ROOT

from .exceptions import DefinitionIDError
from .tables import DATASET_TABLES, SCHEMA_VERSION, glob, load_table


class Captures:
    """Load captures table

    A capture record stores the relationship between a captured file,
    a collection of annotations, and extra metadata that describes this
    capture. For more detail, see schema design here:

    :ref:`captures`

    Examples:

    .. code-block:: python

        >>> captures = Captures(data_root="/data")
        #captures class automatically loads the captures (e.g. lidar scan,
        image, depth map) and the annotations (e.g semantic segmentation
        labels, bounding boxes, etc.)
        >>> data = captures.filter(def_id="6716c783-1c0e-44ae-b1b5-7f068454b66e") # noqa E501 table command not be broken down into multiple lines
        #return the captures and annotations filtered by the annotation
        definition id

    Attributes:
        captures (pd.DataFrame): a collection of captures without annotations
        annotations (pd.DataFrame): a collection of annotations
    """

    TABLE_NAME = "captures"
    FILE_PATTERN = DATASET_TABLES[TABLE_NAME].file

    def __init__(self, data_root=DEFAULT_DATA_ROOT, version=SCHEMA_VERSION):
        """ Initialize Captures

        Args:
            data_root (str): the root directory of the dataset
            version (str): desired schema version
        """
        self.captures = self._load_captures(data_root, version)
        self.annotations = self._load_annotations(data_root, version)

    def _load_captures(self, data_root, version):
        """Load captures except annotations.
        :ref:`captures`

        Args:
            data_root (str): the root directory of the dataset
            version (str): desired schema version

        Returns:
            A pandas dataframe with combined capture records.
            Columns: 'id' (UUID of the capture), 'sequence_id',
            'step' (index of captures), 'timestamp' (Simulation timestamp in
            milliseconds since the sequence started.), 'sensor'
            (sensor attributes), 'ego' (ego pose of the simulation),
            'filename' (single filename that stores captured data)

        Example Captures DataFrame:
         id(str)      sequence_id(str)  step(int)    timestamp(float) \
         cdc8bc5c...  2954c...          300           4.979996

        sensor (dict) \
        {'sensor_id': 'da873b...', 'ego_id': '44ca9...', 'modality': 'camera',
        'translation': [0.0, 0.0, 0.0], 'rotation': [0.0, 0.0, 0.0, 1.0],
        'scale': 0.344577253}


         ego (dict)  \
        {'ego_id': '44ca9...', 'translation': [0.0, 0.0, -20.0],
        'rotation': [0.0, 0.0, 0.0, 1.0], 'velocity': None,
        'acceleration': None}

        filename (str)          format (str)
        RGB3/rgb_30...           PNG

        """
        captures = []
        for c_file in glob(data_root, self.FILE_PATTERN):
            capture = load_table(c_file, self.TABLE_NAME, version, max_level=0)
            if "annotations" in capture.columns:
                capture.drop(columns="annotations")

            captures.append(capture)

        # pd.concat might create memory bottleneck
        return pd.concat(captures, axis=0)

    def _load_annotations(self, data_root, version):
        """Load annotations and capture IDs.
        :ref:`capture-annotation`

        Args:
            data_root (str): the root directory of the dataset
            version (str): desired schema version

        Returns:
            A pandas dataframe with combined annotation records
            Columns: 'id' (annotation id), 'annotation_definition' (annotation
            definition ID),
             'values'
             (list of objects that store annotation data, e.g. 2d bounding
             box), 'capture.id'

        Example Annotation Dataframe:

        id(str)	annotation_definition(str)	\
        ace0...	6716c...

        values (dict) \
        [{'label_id': 34, 'label_name': 'snack_chips_pringles',
        ...'height': 118.0}, {'label_id': 35, '... 'height': 91.0}...]

        capture.id (str)
        cdc8b...
      """
        annotations = []
        for c_file in glob(data_root, self.FILE_PATTERN):
            try:
                annotation = load_table(
                    c_file,
                    self.TABLE_NAME,
                    version,
                    record_path="annotations",
                    meta="id",
                    meta_prefix="capture.",
                )
            except KeyError:
                annotation = pd.DataFrame(
                    {"annotation_definition": [], "capture.id": []}
                )

            annotations.append(annotation)

        return pd.concat(annotations, axis=0)

    def filter(self, def_id):
        """Get captures and annotations filtered by annotation definition id
        :ref:`captures`

        Args:
            def_id (int): annotation definition id used to filter results

        Returns:
            A pandas dataframe with captures and annotations
            Columns: 'id' (capture id), 'sequence_id', 'step', 'timestamp',
            'sensor', 'ego',
             'filename', 'format', 'annotation.id',
             'annotation.annotation_definition','annotation.values'

        Raises:
            DefinitionIDError: Raised if none of the annotation records in the
                combined annotation and captures dataframe match the def_id
                specified as a parameter.

        Example Returned Dataframe (first row):


        +---------------+------------------+-----------+-------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------+------------+---------------+--------------+---------------------+---------------------------------------+-----------------------------------------------------------------------------------------------------------------------+
        | label_id(int) | sequence_id(str) | step(int) | timestamp (float) | sensor (dict)                                                                                                                                                 | ego (dict) | filename(str) | format (str) | annotation.id (str) | annotation.annotation_definition(str) | annotation.values                                                                                                     |
        +===============+==================+===========+===================+===============================================================================================================================================================+============+===============+==============+=====================+=======================================+=======================================================================================================================+
        | 2             | None             | 50        | 4.9               | {'sensor_id': 'dDa873b...', 'ego_id': '44ca9...', 'modality': 'camera','translation': [0.0, 0.0, 0.0], 'rotation': [0.0, 0.0, 0.0, 1.0],'scale': 0.344577253} | ...        | RGB3/asd.png  | PNG          | ace0                | 6716c                                 | [{'label_id': 34, 'label_name': 'snack_chips_pringles',...'height': 118.0}, {'label_id': 35, '... 'height': 91.0}...] |
        +---------------+------------------+-----------+-------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------+------------+---------------+--------------+---------------------+---------------------------------------+-----------------------------------------------------------------------------------------------------------------------+

        """  # noqa: E501 table should not be broken down into multiple lines
        if self.annotations.empty:
            msg = (
                f"Can't find annotations records associate with the given "
                f"definition id {def_id}."
            )
            raise DefinitionIDError(msg)

        mask = self.annotations.annotation_definition == def_id
        annotations = (
            self.annotations[mask]
            .set_index("capture.id")
            .add_prefix("annotation.")
        )
        captures = self.captures.set_index("id")

        combined = (
            captures.join(annotations, how="inner")
            .reset_index()
            .rename(columns={"index": "id"})
        )

        if combined.empty:
            msg = (
                f"Can't find annotations records associate with the given "
                f"definition id {def_id}."
            )
            raise DefinitionIDError(msg)

        return combined



""" Helper keypoints library to plot keypoint joints and skeletons  with a
simple Python API.
"""


def _get_color_from_color_node(color):
    """ Gets the color from the color node in the template.

    Args:
        color (tuple): The color's channel values expressed in a range from 0..1

    Returns: The color for the node.

    """
    r = int(color["r"] * 255)
    g = int(color["g"] * 255)
    b = int(color["b"] * 255)
    a = int(color["a"] * 255)
    return r, g, b, a


def _get_color_for_bone(bone):
    """ Gets the color for the bone from the template. A bone is a visual
        connection between two keypoints in the keypoint list of the figure.

        bone
        {
            joint1: <int> Index into the keypoint list for the first joint.
            joint2: <int> Index into the keypoint list for the second joint.
            color {
                r: <float> Value (0..1) of the red channel.
                g: <float> Value (0..1) of the green channel.
                b: <float> Value (0..1) of the blue channel.
                a: <float> Value (0..1) of the alpha channel.
            }
        }

    Args:
        bone: The active bone.

    Returns: The color of the bone.

    """
    if "color" in bone:
        return _get_color_from_color_node(bone["color"])
    else:
        return 255, 0, 255, 255


def _get_color_for_keypoint(template, keypoint):
    """ Gets the color for the keypoint from the template. A keypoint is a
        location of interest inside of a figure. Keypoints are connected
        together with bones. The configuration of keypoint locations and bone
        connections are defined in a template file.

    keypoint_template {
        template_id: <str> The UUID of the template.
        template_name: <str> Human readable name of the template.
        key_points [ <List> List of joints defined in this template
            {
                label: <str> The label of the joint.
                index: <int> The index of the joint.
                color {
                    r: <float> Value (0..1) for the red channel.
                    g: <float> Value (0..1) for the green channel.
                    b: <float> Value (0..1) for the blue channel.
                    a: <float> Value (0..1) for the alpha channel.
                }
            }, ...
        ]
        skeleton [ <List> List of skeletal connections
            {
                joint1: <int> The first joint of the connection.
                joint2: <int> The second joint of the connection.
                color {
                    r: <float> Value (0..1) for the red channel.
                    g: <float> Value (0..1) for the green channel.
                    b: <float> Value (0..1) for the blue channel.
                    a: <float> Value (0..1) for the alpha channel.
                }
            }, ...
        ]
    }

    Args:
        template: The active template.
        keypoint: The active keypoint.

    Returns: The color for the keypoint.

    """
    node = template["key_points"][keypoint["index"]]

    if "color" in node:
        return _get_color_from_color_node(node["color"])
    else:
        return 0, 0, 255, 255


def draw_keypoints_for_figure(image, figure, draw, templates, visual_width=6):
    """ Draws keypoints for a figure on an image.

    keypoints {
        label_id: <int> Integer identifier of the label.
        instance_id: <str> UUID of the instance.
        template_guid: <str> UUID of the keypoint template.
        pose: <str> String label for current pose.
        keypoints [
            {
                index: <int> Index of keypoint in template.
                x: <float> X subpixel coordinate of keypoint.
                y: <float> Y subpixel coordinate of keypoint
                state: <int> 0: keypoint does not exist,
                             1: keypoint exists but is not visible,
                             2: keypoint exists and is visible.
            }, ...
        ]
    }

    Args:
        image (PIL Image): a PIL image.
        figure: The figure to draw.
        draw (PIL ImageDraw): PIL image draw interface.
        templates (list): a list of keypoint templates.
        visual_width (int): the visual width of the joints.

    Returns: a PIL image with keypoints for a figure drawn on it.

    """
    # find the template for this
    for template in templates:
        if template["template_id"] == figure["template_guid"]:
            break
    else:
        return image

    # load the spec
    skeleton = template["skeleton"]

    for bone in skeleton:
        j1 = figure["keypoints"][bone["joint1"]]
        j2 = figure["keypoints"][bone["joint2"]]

        if j1["state"] == 2 and j2["state"] == 2:
            x1 = int(j1["x"])
            y1 = int(j1["y"])
            x2 = int(j2["x"])
            y2 = int(j2["y"])

            color = _get_color_for_bone(bone)
            draw.line((x1, y1, x2, y2), fill=color, width=visual_width)

    for k in figure["keypoints"]:
        state = k["state"]
        if state == 2:
            x = k["x"]
            y = k["y"]

            color = _get_color_for_keypoint(template, k)

            half_width = visual_width / 2

            draw.ellipse(
                (
                    x - half_width,
                    y - half_width,
                    x + half_width,
                    y + half_width,
                ),
                fill=color,
                outline=color,
            )

    return image



import hashlib
import logging
import os
import re
import tempfile
import zlib
from pathlib import Path

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from .exceptions import ChecksumError, DownloadError

logger = logging.getLogger(__name__)

# Timeout of requests (in seconds)
DEFAULT_TIMEOUT = 1800
# Retry after failed request
DEFAULT_MAX_RETRIES = 5


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, timeout, *args, **kwargs):
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def download_file(source_uri: str, dest_path: str, file_name: str = None):
    """Download a file specified from a source uri

    Args:
        source_uri (str): source url where the file should be downloaded
        dest_path (str): destination path of the file
        file_name (str): file name of the file to be downloaded

    Returns:
        String of destination path.
    """
    logger.debug(f"Trying to download file from {source_uri} -> {dest_path}")
    adapter = TimeoutHTTPAdapter(
        timeout=DEFAULT_TIMEOUT, max_retries=Retry(total=DEFAULT_MAX_RETRIES)
    )
    with requests.Session() as http:
        http.mount("https://", adapter)
        try:
            response = http.get(source_uri)
            response.raise_for_status()
        except requests.exceptions.RequestException as ex:
            logger.error(ex)
            err_msg = (
                f"The request download from {source_uri} -> {dest_path} can't "
                f"be completed."
            )

            raise DownloadError(err_msg)
        else:
            dest_path = Path(dest_path)
            if not file_name:
                file_name = _parse_filename(response, source_uri)
            dest_path = dest_path / file_name
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            with open(dest_path, "wb") as f:
                f.write(response.content)

    return dest_path


def checksum_matches(filepath, expected_checksum, algorithm="CRC32"):
    """ Check if the checksum matches

    Args:
        filepath (str): the doaloaded file path
        expected_checksum (int): expected checksum of the file
        algorithm (str): checksum algorithm. Defaults to CRC32

    Returns:
        True if the file checksum matches.
    """
    computed = compute_checksum(filepath, algorithm)
    return computed == expected_checksum


def validate_checksum(filepath, expected_checksum, algorithm="CRC32"):
    """ Validate checksum of the downloaded file.

    Args:
        filepath (str): the doaloaded file path
        expected_checksum (int): expected checksum of the file
        algorithm (str): checksum algorithm. Defaults to CRC32

    Raises:
        ChecksumError if the file checksum does not match.
    """
    if not checksum_matches(filepath, expected_checksum, algorithm):
        raise ChecksumError


def compute_checksum(filepath, algorithm="CRC32"):
    """ Compute the checksum of a file.

    Args:
        filepath (str): the doaloaded file path
        algorithm (str): checksum algorithm. Defaults to CRC32

    Returns:
        int: the checksum value
    """
    if algorithm == "CRC32":
        chs = _crc32_checksum(filepath)
    elif algorithm == "MD5":
        chs = _md5_checksum(filepath)
    else:
        raise ValueError("Unsupported checksum algorithm!")

    return chs


def _crc32_checksum(filepath):
    """ Calculate the checksum of a file using CRC32.
    """
    with open(filepath, "rb") as f:
        checksum = zlib.crc32(f.read())

    return checksum


def _md5_checksum(filename):
    """ Calculate the checksum of a file using MD5.
    """
    md5 = hashlib.md5()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            md5.update(chunk)
    return md5.hexdigest()


def get_checksum_from_file(filepath):
    """ This method return checksum of the file whose filepath is given.

    Args:
        filepath (str): Path of the checksum file.
                        Path can be HTTP(s) url or local path.

    Raises:
        ValueError: Raises this error if filepath is not local or not
                    HTTP or HTTPS url.

    """

    if filepath.startswith(("http://", "https://")):
        with tempfile.TemporaryDirectory() as tmp:
            checksum_file_path = os.path.join(tmp, "checksum.txt")
            file_path = download_file(
                source_uri=filepath, dest_path=checksum_file_path
            )
            return _read_checksum_from_txt(file_path)

    elif os.path.isfile(filepath):
        return _read_checksum_from_txt(filepath)

    else:
        raise ValueError(f"Can not get checksum from path: {filepath}")


def _read_checksum_from_txt(filepath):
    """ This method reads checksum from a txt file and returns it.

    Args:
        filepath (str): Local filepath of the checksum file.

    Returns:
        str: checksum value from the checksum file.

    """
    with open(filepath) as file:
        checksum = file.read()
    return checksum


def _parse_filename(response, uri):
    file_name = _get_filename_from_response(response)
    if file_name is None:
        file_name = _get_file_name_from_uri(uri)
    return file_name


def _get_filename_from_response(response):
    """ Gets filename from requests response object

        Args:
            response: requests.Response() object that contains the server's
            response to the HTTP request.

        Returns:
            filename (str): Name of the file to be downloaded
    """
    cd = response.headers.get("content-disposition")
    if not cd:
        return None
    file_name = re.findall("filename=(.+)", cd)
    if len(file_name) == 0:
        return None
    return file_name[0]


def _get_file_name_from_uri(uri):
    """ Gets filename from URI

    Args:
        uri (str): URI

    """
    return uri.split("/")[-1]



DEFAULT_DATA_ROOT = "/data"

# Default Unity Project ID where USim jobs was executed
DEFAULT_PROJECT_ID = "474ba200-4dcc-4976-818e-0efd28efed30"
USIM_API_ENDPOINT = "https://api.simulation.unity3d.com"

# Default Timing text for codetiming.Timer decorator
TIMING_TEXT = "[{name}] elapsed time: {:0.4f} seconds."

# Click CLI context settings
CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
    "show_default": True,
    "ignore_unknown_options": True,
    "allow_extra_args": True,
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datasetinsights',
 'datasetinsights.commands',
 'datasetinsights.datasets',
 'datasetinsights.datasets.unity_perception',
 'datasetinsights.io',
 'datasetinsights.io.downloader',
 'datasetinsights.stats',
 'datasetinsights.stats.visualization']

package_data = \
{'': ['*'], 'datasetinsights.stats.visualization': ['font/*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'codetiming>=1.2.0,<2.0.0',
 'cython>=0.29.14,<0.30.0',
 'dash==1.12.0',
 'dask[complete]>=2.14.0,<3.0.0',
 'google-cloud-storage>=1.24.1,<2.0.0',
 'jupyter>=1.0.0,<2.0.0',
 'matplotlib>=3.3.1,<4.0.0',
 'numpy>=1.17,<2.0',
 'opencv-python>=4.4.0.42,<5.0.0.0',
 'pandas>=1.0.1,<2.0.0',
 'plotly>=4.4.1,<5.0.0',
 'pyquaternion>=0.9.5,<0.10.0',
 'tqdm>=4.45.0,<5.0.0']

entry_points = \
{'console_scripts': ['datasetinsights = datasetinsights.__main__:entrypoint']}

setup_kwargs = {
    'name': 'datasetinsights',
    'version': '1.1.0',
    'description': 'Synthetic dataset insights.',
    'long_description': '# Dataset Insights\n\n[![PyPI python](https://img.shields.io/pypi/pyversions/datasetinsights)](https://pypi.org/project/datasetinsights)\n[![PyPI version](https://badge.fury.io/py/datasetinsights.svg)](https://pypi.org/project/datasetinsights)\n[![Downloads](https://pepy.tech/badge/datasetinsights)](https://pepy.tech/project/datasetinsights)\n[![Tests](https://github.com/Unity-Technologies/datasetinsights/actions/workflows/linting-and-unittests.yaml/badge.svg?branch=master&event=push)](https://github.com/Unity-Technologies/datasetinsights/actions/workflows/linting-and-unittests.yaml?query=branch%3Amaster+event%3Apush)\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](LICENSE)\n\nUnity Dataset Insights is a python package for downloading, parsing and analyzing synthetic datasets generated using the Unity [Perception package](https://github.com/Unity-Technologies/com.unity.perception).\n\n## Installation\n\nDataset Insights maintains a pip package for easy installation. It can work in any standard Python environment using `pip install datasetinsights` command.\n\n## Getting Started\n\n### Dataset Statistics\n\nWe provide a sample [notebook](notebooks/Perception_Statistics.ipynb) to help you load synthetic datasets generated using [Perception package](https://github.com/Unity-Technologies/com.unity.perception) and visualize dataset statistics. We plan to support other sample Unity projects in the future.\n\n### Dataset Download\n\nYou can download the datasets from HTTP(s), GCS, and Unity simulation projects using the \'download\' command from CLI or API.\n\n[CLI](https://datasetinsights.readthedocs.io/en/latest/datasetinsights.commands.html#datasetinsights-commands-download)\n\n```bash\ndatasetinsights download \\\n  --source-uri=<xxx> \\\n  --output=$HOME/data\n```\n[Programmatically](https://datasetinsights.readthedocs.io/en/latest/datasetinsights.io.downloader.html#module-datasetinsights.io.downloader.gcs_downloader)\n\nUnitySimulationDownloader downloads a dataset from Unity Simulation.\n\n```python3\nfrom datasetinsights.io.downloader import UnitySimulationDownloader\n\nsource_uri=usim://<project_id>/<run_execution_id>\ndest = "~/data"\naccess_token = "XXX"\ndownloader = UnitySimulationDownloader(access_token=access_token)\ndownloader.download(source_uri=source_uri, output=dest)\n```\nGCSDatasetDownloader downloads a dataset from GCS location.\n```python3\nfrom datasetinsights.io.downloader import GCSDatasetDownloader\n\nsource_uri=gs://url/to/file.zip or gs://url/to/folder\ndest = "~/data"\ndownloader = GCSDatasetDownloader()\ndownloader.download(source_uri=source_uri, output=dest)\n```\nHTTPDatasetDownloader downloads a dataset from any HTTP(S) location.\n```python3\nfrom datasetinsights.io.downloader import HTTPDatasetDownloader\n\nsource_uri=http://url.to.file.zip\ndest = "~/data"\ndownloader = HTTPDatasetDownloader()\ndownloader.download(source_uri=source_uri, output=dest)\n```\n### Dataset Explore\nYou can explore the dataset [schema](https://datasetinsights.readthedocs.io/en/latest/Synthetic_Dataset_Schema.html#synthetic-dataset-schema) by using following API:\n\n[Unity Perception](https://datasetinsights.readthedocs.io/en/latest/datasetinsights.datasets.unity_perception.html#datasetinsights-datasets-unity-perception)\n\nAnnotationDefinitions and MetricDefinitions loads synthetic dataset definition tables and return a dictionary containing the definitions.\n\n```python3\nfrom datasetinsights.datasets.unity_perception import AnnotationDefinitions,\nMetricDefinitions\nannotation_def = AnnotationDefinitions(data_root=dest, version="my_schema_version")\ndefinition_dict = annotation_def.get_definition(def_id="my_definition_id")\n\nmetric_def = MetricDefinitions(data_root=dest, version="my_schema_version")\ndefinition_dict = metric_def.get_definition(def_id="my_definition_id")\n```\nCaptures loads synthetic dataset captures tables and return a pandas dataframe with captures and annotations columns.\n\n```python3\nfrom datasetinsights.datasets.unity_perception import Captures\ncaptures = Captures(data_root=dest, version="my_schema_version")\ncaptures_df = captures.filter(def_id="my_definition_id")\n```\nMetrics loads synthetic dataset metrics table which holds extra metadata that can be used to describe a particular sequence, capture or annotation and return a pandas dataframe with captures and metrics columns.\n\n```python3\nfrom datasetinsights.datasets.unity_perception import Metrics\nmetrics = Metrics(data_root=dest, version="my_schema_version")\nmetrics_df = metrics.filter_metrics(def_id="my_definition_id")\n```\n\n## Docker\n\nYou can use the pre-build docker image [unitytechnologies/datasetinsights](https://hub.docker.com/r/unitytechnologies/datasetinsights) to run similar commands.\n\n## Documentation\n\nYou can find the API documentation on [readthedocs](https://datasetinsights.readthedocs.io/en/latest/).\n\n## Contributing\n\nPlease let us know if you encounter a bug by filing an issue. To learn more about making a contribution to Dataset Insights, please see our Contribution [page](CONTRIBUTING.md).\n\n## License\n\nDataset Insights is licensed under the Apache License, Version 2.0. See [LICENSE](LICENCE) for the full license text.\n\n## Citation\nIf you find this package useful, consider citing it using:\n```\n@misc{datasetinsights2020,\n    title={Unity {D}ataset {I}nsights Package},\n    author={{Unity Technologies}},\n    howpublished={\\url{https://github.com/Unity-Technologies/datasetinsights}},\n    year={2020}\n}\n```\n',
    'author': 'Unity AI Perception Team',
    'author_email': 'computer-vision@unity3d.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Unity-Technologies/datasetinsights',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from PIL import Image, ImageColor, ImageDraw

from datasetinsights.stats.visualization.bbox2d_plot import (
    add_single_bbox_on_image,
)
from datasetinsights.stats.visualization.bbox3d_plot import (
    add_single_bbox3d_on_image,
)

from .keypoints_plot import draw_keypoints_for_figure

logger = logging.getLogger(__name__)
COLORS = list(ImageColor.colormap.values())
FONT_SCALE = 35
LINE_WIDTH_SCALE = 250
ERROR_BAR_BASE_COLOR = "indianred"
ERROR_BAR_COMPARE_COLOR = "lightseagreen"


def grid_plot(images, figsize=(3, 5), img_type="rgb", titles=None):
    """ Plot 2D array of images in grid.
    Args:
        images (list): 2D array of images.
        figsize (tuple): target figure size of each image in the grid.
        Defaults to (3, 5).
        img_type (string): image plot type ("rgb", "gray"). Defaults to "rgb".
        titles (list[str]): a list of titles. Defaults to None.
    Returns:
        matplotlib figure the combined grid plot.
    """
    n_rows = len(images)
    n_cols = len(images[0])

    figsize = (figsize[0] * n_cols, figsize[1] * n_rows)

    figure = plt.figure(figsize=figsize, constrained_layout=True)
    for i in range(n_rows):
        for j in range(n_cols):
            img = images[i][j]
            k = i * n_cols + j + 1
            plt.subplot(n_rows, n_cols, k)
            plt.xticks([])
            plt.yticks([])
            if titles:
                plt.title(titles[k - 1])
            plt.grid(False)
            if img_type == "gray":
                plt.imshow(img, cmap="gray")
            else:
                plt.imshow(img, plt.cm.binary)
    plt.show()

    return figure


def _process_label(bbox, label_mappings=None):
    """Create a label text for the bbox.

    Args:
        bbox (BBox2D): a bounding box
        label_mappings (dict): a dict of {label_id: label_name} mapping
        Defaults to None.
    """
    if label_mappings is not None:
        label = label_mappings[bbox.label]
    else:
        label = bbox.label
    if bbox.score != 1.0:
        return f"{label}: {bbox.score * 100: .2f}%"
    else:
        return label


def plot_bboxes3d(image, bboxes, projection, colors=None, orthographic=False):
    """ Plot an image with 3D bounding boxes

    Currently this method should only be used for ground truth images, and
    doesn't support predictions. If a list of colors is not provided as an
    argument to this routine, the default color of green will be used.

    Args:
        image (PIL Image): a PIL image.
        bboxes (list): a list of BBox3D objects
        projection: The perspective projection of the camera which
        captured the ground truth.
        colors (list): a color list for boxes. Defaults to none. If
        colors = None, it will default to coloring all boxes green.
        orthographic (bool): true if proj is orthographic, else perspective

    Returns:
        PIL image: a PIL image with bounding boxes drawn on it.
    """
    np_image = np.array(image)
    img_height, img_width, _ = np_image.shape

    for i, box in enumerate(bboxes):
        color = colors[i] if colors else None
        add_single_bbox3d_on_image(
            np_image, box, projection, color, orthographic=orthographic
        )

    return Image.fromarray(np_image)


def plot_bboxes(image, bboxes, label_mappings=None, colors=None):
    """ Plot an image with bounding boxes.

    For ground truth image, a color is randomly selected for each bounding box.
    For prediction, the color of a boundnig box is coded based on IOU value
    between prediction and ground truth bounding boxes. It is considered true
    positive if IOU >= 0.5. We only visualize prediction bounding box with
    score >= 0.5. For prediction, it's a green box if the predicted bounding box
    can be matched to a ground truth bounding boxes. It's a red box if the
    predicted bounding box can't be matched to a ground truth bounding boxes.

    Args:
        image (PIL Image): a PIL image.
        bboxes (list): a list of BBox2D objects.
        label_mappings (dict): a dict of {label_id: label_name} mapping
        Defaults to None.
        colors (list): a color list for boxes. Defaults to None.
        If colors = None, it will randomly assign PIL.COLORS for each box.

    Returns:
        PIL Image: a PIL image with bounding boxes drawn.
    """
    np_image = np.array(image)
    image_height, _, _ = np_image.shape
    for i, box in enumerate(bboxes):
        label = _process_label(box, label_mappings)
        color = colors[i] if colors else None
        font_size = image_height // FONT_SCALE
        box_line_width = image_height // LINE_WIDTH_SCALE
        add_single_bbox_on_image(
            np_image,
            box,
            label,
            color,
            font_size=font_size,
            box_line_width=box_line_width,
        )

    return Image.fromarray(np_image)


def plot_keypoints(image, annotations, templates, visual_width=6):
    """ Plot an image with keypoint data.

    Currently only used for ground truth info. Keypoints and colors are defined
    in templates.

    Args:
        image (PIL Image): a PIL image.
        annotations (list): a list of keypoint annotation data.
        templates (list): a list of keypoint templates.
        visual_width (int): the width of the visual elements

    Returns:
        PIL Image: a PIL image with keypoints drawn.
    """
    draw = ImageDraw.Draw(image)

    for figure in annotations:
        draw_keypoints_for_figure(image, figure, draw, templates, visual_width)

    return image


def bar_plot(
    df, x, y, title=None, x_title=None, y_title=None, x_tickangle=0, **kwargs
):
    """Create plotly bar plot

    Args:
        df (pd.DataFrame): A pandas dataframe that contain bar plot data.
        x (str): The column name of the data in x-axis.
        y (str): The column name of the data in y-axis.
        title (str, optional): The title of this plot.
        x_title (str, optional): The x-axis title.
        y_title (str, optional): The y-axis title.
        x_tickangle (int, optional): X-axis text tickangle (default: 0)

    This method can also take addition keyword arguments that can be passed to
    [plotly.express.bar](https://plotly.com/python-api-reference/generated/plotly.express.bar.html#plotly.express.bar) method.

    Examples:
        >>> import pandas as pd
        >>> df = pd.DataFrame({"id": [0, 1, 2], "name": ["a", "b", "c"],
        ...                    "count": [10, 20, 30]})
        >>> bar_plot(df, x="id", y="count", hover_name="name")
    """  # noqa: E501 URL should not be broken down into lines
    fig = px.bar(df, x=x, y=y, **kwargs)
    fig = fig.update_layout(
        xaxis=dict(title=x_title, tickangle=x_tickangle),
        yaxis=dict(title=y_title),
        title_text=title,
    )

    return fig


def histogram_plot(
    df, x, max_samples=None, title=None, x_title=None, y_title=None, **kwargs
):
    """Create plotly histogram plot

    Args:
        df (pd.DataFrame): A pandas dataframe that contain raw data.
        x (str): The column name of the raw data for histogram plot.
        title (str, optional): The title of this plot.
        x_title (str, optional): The x-axis title.
        y_title (str, optional): The y-axis title.

    This method can also take addition keyword arguments that can be passed to
    [plotly.express.histogram](https://plotly.com/python-api-reference/generated/plotly.express.histogram.html) method.

    Examples:
        >>> import pandas as pd
        >>> df = pd.DataFrame({"id": [0, 1, 2], "count": [10, 20, 30]})
        >>> histogram_plot(df, x="count")

        Histnorm plot using probability density:

        >>> histogram_plot(df, x="count", histnorm="probability density")
    """  # noqa: E501 URL should not be broken down into lines
    if max_samples and len(df) > max_samples:
        df = df.sample(n=max_samples)

    fig = px.histogram(df, x=x, **kwargs)
    fig = fig.update_layout(
        xaxis=dict(title=x_title), yaxis=dict(title=y_title), title_text=title,
    )

    return fig


def _convert_euler_rotations_to_scatter_points(
    df, x_col, y_col, z_col=None, max_samples=None
):
    """Turns euler rotations into a dataframe of points for plotting."""
    if max_samples is not None and len(df) > max_samples:
        df = df.sample(max_samples)

    for _, row in df.iterrows():
        v1 = np.array((0, 1, 0))

        theta_x = np.radians(row[x_col])

        rx = np.array(
            (
                (np.cos(theta_x), -np.sin(theta_x), 0),
                (np.sin(theta_x), np.cos(theta_x), 0),
                (0, 0, 1),
            )
        )
        v1 = rx.dot(v1)

        theta_y = np.radians(row[y_col])
        ry = np.array(
            (
                (1, 0, 0),
                (0, np.cos(theta_y), -np.sin(theta_y)),
                (0, np.sin(theta_y), np.cos(theta_y)),
            )
        )
        v1 = ry.dot(v1)

        if z_col is not None:
            theta_z = np.radians(row[z_col])
            rz = np.array(
                (
                    (np.cos(theta_z), 0, np.sin(theta_z)),
                    (0, 1, 0),
                    (-np.sin(theta_z), 0, np.cos(theta_z)),
                )
            )
            v1 = rz.dot(v1)

            text = """x: {x}°  y: {y}°  z: {z}°""".format(
                x=row[x_col], y=row[y_col], z=row[z_col]
            )
        else:
            text = """x: {x}°  y: {y}°""".format(x=row[x_col], y=row[y_col])

        ser = [v1[0], v1[1], v1[2], text]
        yield ser


def rotation_plot(df, x, y, z=None, max_samples=None, title=None, **kwargs):
    """Create a plotly 3d rotation plot
    Args:
        df (pd.DataFrame): A pandas dataframe that contains the raw data.
        x (str): The column name containing x rotations.
        y (str): The column name containing y rotations.
        z (str, optional): The column name containing z rotations.
        title (str, optional): The title of this plot.

    This method can also take addition keyword arguments that can be passed to
    [plotly.graph_objects.Scatter3d](https://plotly.com/python-api-reference/generated/plotly.graph_objects.Scatter3d.html) method.

    Returns:
        A plotly.graph_objects.Figure containing the scatter plot
    """  # noqa: E501 URL should not be broken down into lines
    rot_plot_columns = ("x", "y", "z", "text")
    dfrot = pd.DataFrame(
        list(
            _convert_euler_rotations_to_scatter_points(df, x, y, z, max_samples)
        ),
        columns=rot_plot_columns,
    )
    fig = (
        go.Figure(
            data=[
                go.Scatter3d(
                    x=dfrot["x"],
                    y=dfrot["y"],
                    z=dfrot["z"],
                    text=dfrot["text"],
                    mode="markers",
                    hoverinfo="text",
                    marker=dict(size=5, opacity=0.5),
                    **kwargs,
                )
            ]
        )
        .update_xaxes(showticklabels=False)
        .update_layout(
            title_text=title,
            scene=dict(
                xaxis=dict(showticklabels=False, range=[-1, 1]),
                yaxis=dict(showticklabels=False, range=[-1, 1]),
                zaxis=dict(showticklabels=False, range=[-1, 1]),
            ),
        )
    )
    return fig


def model_performance_comparison_box_plot(
    title=None,
    mean_ap_base=None,
    mean_ap_50_base=None,
    mean_ar_base=None,
    mean_ap_new=None,
    mean_ap_50_new=None,
    mean_ar_new=None,
    range=[0, 1.0],
    **kwargs,
):
    """Create a box plot for a base and new model performance
    Args:
        title (str): title of the plot
        mean_ap_base (list): a list of base mAP
        mean_ap_50_base (list): a list of base mAP
        mean_ar_base (list): a list of base mAP
        mean_ap_new (list): a list of base mAP
        mean_ap_50_new (list): a list of base mAP
        mean_ar_new (list): a list of base mAP
        range (list): the range of y axis. Defaults to [0, 1.0]

    Returns:
        A plotly.graph_objects.Figure containing the box plot
    """
    fig = go.Figure(
        layout=go.Layout(title=go.layout.Title(text=title), **kwargs)
    )
    fig.update_yaxes(range=range)
    _fig_add_trace(
        fig,
        mean_ap_base,
        name="baes mAP",
        base=True,
        color=ERROR_BAR_BASE_COLOR,
    )
    _fig_add_trace(
        fig,
        mean_ap_new,
        name="new mAP",
        base=False,
        color=ERROR_BAR_COMPARE_COLOR,
    )
    _fig_add_trace(
        fig,
        mean_ap_50_base,
        name="base mAP50",
        base=True,
        color=ERROR_BAR_BASE_COLOR,
    )
    _fig_add_trace(
        fig,
        mean_ap_50_new,
        name="new mAP50",
        base=False,
        color=ERROR_BAR_COMPARE_COLOR,
    )
    _fig_add_trace(
        fig,
        mean_ar_base,
        name="base mAR",
        base=True,
        color=ERROR_BAR_BASE_COLOR,
    )
    _fig_add_trace(
        fig,
        mean_ar_new,
        name="new mAR",
        base=False,
        color=ERROR_BAR_COMPARE_COLOR,
    )

    return fig


def model_performance_box_plot(
    title=None,
    mean_ap=None,
    mean_ap_50=None,
    mean_ar=None,
    range=[0, 1.0],
    **kwargs,
):
    """Create a box plot for one model performance
    Args:
        title (str): title of the plot
        mean_ap (list): a list of base mAP
        mean_ap_50 (list): a list of base mAP
        mean_ar (list): a list of base mAP
        range (list): the range of y axis. Defaults to [0, 1.0]

    Returns:
        A plotly.graph_objects.Figure containing the box plot
    """
    fig = go.Figure(
        layout=go.Layout(title=go.layout.Title(text=title), **kwargs)
    )
    fig.update_yaxes(range=range)
    _fig_add_trace(fig, mean_ap, name="mAP")
    _fig_add_trace(fig, mean_ap_50, name="mAP@IOU50")
    _fig_add_trace(fig, mean_ar, name="mAR")

    return fig


def _fig_add_trace(
    fig=None, data=[], name="", base=True, color=ERROR_BAR_BASE_COLOR
):
    """Add box plot in figure
    Args:
        fig (go.Figure): figure you want to add box plot in
        data (list): metric values
        name (str): name of the added box plot
        base (bool): whether is a base metric. Defaults to true
        color (str): color for box plot

    Returns:
        A plotly.graph_objects.Figure containing the box plot
    """
    if not data:
        return
    fig.add_trace(go.Box(y=data, name=name, marker_color=color))



""" Helper bounding box 3d library to plot pretty 3D boundign
boxes with a simple Python API.
"""

import cv2
import numpy


def _add_single_bbox3d_on_image(
    image,
    front_bottom_left,
    front_upper_left,
    front_upper_right,
    front_bottom_right,
    back_bottom_left,
    back_upper_left,
    back_upper_right,
    back_bottom_right,
    color=None,
    box_line_width=2,
):
    """ Add a single 3D bounding box to the passed in image.

    For this version of the method, all of the passed in coordinates should be
    integer tuples already projected in image pixel coordinate space.

    Args:
        image (numpy array): numpy array version of the image
        front_bottom_left (int tuple): Front bottom left coordinate of the 3D
        bounding box in pixel space
        front_upper_left (int tuple): Front upper left coordinate of the 3D
        bounding box in pixel space
        front_upper_right (int tuple): Front upper right coordinate of the 3D
        bounding box in pixel space
        front_bottom_right (int tuple): Front bottom right coordinate of the 3D
        bounding box in pixel space
        back_bottom_left (int tuple): Back bottom left coordinate of the 3D
        bounding box in pixel space
        back_upper_left (int tuple): Back bottom left coordinate of the 3D
        bounding box in pixel space
        back_upper_right (int tuple): Back bottom left coordinate of the 3D
        bounding box in pixel space
        back_bottom_right (int tuple): Back bottom left coordinate of the 3D
        bounding box in pixel space
        color (tuple): RGBA color of the bounding box. Defaults to None. If
        color = None the the tuple of [0, 255, 0, 255] (Green) will be used.
        box_line_width: The width of the drawn box. Defaults to 2.
    """
    try:
        fbl = (front_bottom_left[0], front_bottom_left[1])
        ful = (front_upper_left[0], front_upper_left[1])
        fur = (front_upper_right[0], front_upper_right[1])
        fbr = (front_bottom_right[0], front_bottom_right[1])

        bbl = (back_bottom_left[0], back_bottom_left[1])
        bul = (back_upper_left[0], back_upper_left[1])
        bur = (back_upper_right[0], back_upper_right[1])
        bbr = (back_bottom_right[0], back_bottom_right[1])

    except ValueError:
        raise TypeError("all box coorinates must be a number")

    if color is None:
        color = [0, 255, 0, 255]

    cv2.line(image, fbl, ful, color, box_line_width)  # front left
    cv2.line(image, ful, fur, color, box_line_width)  # front top
    cv2.line(image, fbr, fur, color, box_line_width)  # front right
    cv2.line(image, fbl, fbr, color, box_line_width)  # front bottom

    cv2.line(image, bbl, bul, color, box_line_width)  # back left
    cv2.line(image, bul, bur, color, box_line_width)  # back top
    cv2.line(image, bbr, bur, color, box_line_width)  # back right
    cv2.line(image, bbl, bbr, color, box_line_width)  # back bottom

    cv2.line(image, ful, bul, color, box_line_width)  # top left
    cv2.line(image, fur, bur, color, box_line_width)  # top right
    cv2.line(image, fbl, bbl, color, box_line_width)  # bottom left
    cv2.line(image, fbr, bbr, color, box_line_width)  # bottom right


def add_single_bbox3d_on_image(
    image, box, proj, color=None, box_line_width=2, orthographic=False,
):
    """" Add single 3D bounding box on a given image.

    Args:
        image (numpy array): a numpy array for an image
        box (BBox3D): a 3D bounding box in camera's coordinate system
        proj (numpy 2D array): camera's 3x3 projection matrix
        color(tuple): RGBA color of the bounding box. Defaults to None. If
        color = None the the tuple of [0, 255, 0, 255] (Green) will be used.
        box_line_width (int): line width of the bounding boxes. Defaults to 2.
        orthographic (bool): true if proj is orthographic, else perspective
    """
    img_height, img_width, _ = image.shape

    fll = box.back_left_bottom_pt
    ful = box.back_left_top_pt
    fur = box.back_right_top_pt
    flr = box.back_right_bottom_pt

    bll = box.front_left_bottom_pt
    bul = box.front_left_top_pt
    bur = box.front_right_top_pt
    blr = box.front_right_bottom_pt

    project_pt_to_pixel = (
        _project_pt_to_pixel_location_orthographic
        if orthographic
        else _project_pt_to_pixel_location
    )

    fll_raster = project_pt_to_pixel(fll, proj, img_height, img_width)
    ful_raster = project_pt_to_pixel(ful, proj, img_height, img_width)
    fur_raster = project_pt_to_pixel(fur, proj, img_height, img_width)
    flr_raster = project_pt_to_pixel(flr, proj, img_height, img_width)
    bll_raster = project_pt_to_pixel(bll, proj, img_height, img_width)
    bul_raster = project_pt_to_pixel(bul, proj, img_height, img_width)
    bur_raster = project_pt_to_pixel(bur, proj, img_height, img_width)
    blr_raster = project_pt_to_pixel(blr, proj, img_height, img_width)

    _add_single_bbox3d_on_image(
        image,
        fll_raster,
        ful_raster,
        fur_raster,
        flr_raster,
        bll_raster,
        bul_raster,
        bur_raster,
        blr_raster,
        color,
        box_line_width,
    )


def _project_pt_to_pixel_location(pt, projection, img_height, img_width):
    """ Projects a 3D coordinate into a pixel location from a perspective camera.

    Applies the passed in projection matrix to project a point from the camera's
    coordinate space into pixel space.

    For a description of the math used in this method, see:
    https://www.scratchapixel.com/lessons/3d-basic-rendering/computing-pixel-coordinates-of-3d-point/

    Args:
        pt (numpy array): The 3D point to project.
        projection (numpy 2D array): The camera's 3x3 projection matrix.
        img_height (int): The height of the image in pixels.
        img_width (int): The width of the image in pixels.

    Returns:
        numpy array: a one-dimensional array with two values (x and y)
        representing a point's pixel coordinate in an image.
    """

    _pt = projection.dot(pt)

    # compute the perspective divide. Near clipping plane should take care of
    # divide by zero cases, but we will check to be sure
    if _pt[2] != 0:
        _pt /= _pt[2]

    return numpy.array(
        [
            int(-(_pt[0] * img_width) / 2.0 + (img_width * 0.5)),
            int((_pt[1] * img_height) / 2.0 + (img_height * 0.5)),
        ]
    )


def _project_pt_to_pixel_location_orthographic(
    pt, projection, img_height, img_width
):
    """ Projects a 3D coordinate into a pixel location from an orthographic camera.

        Applies the passed in projection matrix to project a point from the
        camera's coordinate space into pixel space.

        For a description of the math used in this method, see:
        https://www.scratchapixel.com/lessons/3d-basic-rendering/perspective-and-
        orthographic-projection-matrix/projection-matrix-introduction

        Args:
            pt (numpy array): The 3D point to project.
            projection (numpy 2D array): The camera's 3x3 projection matrix.
            img_height (int): The height of the image in pixels.
            img_width (int): The width of the image in pixels.

        Returns:
            numpy array: a one-dimensional array with two values (x and y)
            representing a point's pixel coordinate in an image.
    """

    # The 'y' component needs to be flipped because of how Unity works
    projection = numpy.array(
        [
            [projection[0][0], 0, 0],
            [0, -projection[1][1], 0],
            [0, 0, projection[2][2]],
        ]
    )
    temp = projection.dot(pt)

    pixel = [
        int((temp[0] + 1) * 0.5 * img_width),
        int((temp[1] + 1) * 0.5 * img_height),
    ]
    return pixel



import logging
import re

import click

import datasetinsights.constants as const
from datasetinsights.io.downloader.base import create_dataset_downloader

logger = logging.getLogger(__name__)


class SourceURI(click.ParamType):
    """Represents the Source URI Parameter type.

    This extends click.ParamType that allows click framework to validates
    supported source URI according to the prefix pattern.

    Raises:
        click.BadParameter: if the validation failed.
    """

    name = "source_uri"
    PREFIX_PATTERN = r"^gs://|^http(s)?://|^usim://"

    def convert(self, value, param, ctx):
        """ Validate source URI and Converts the value.
        """
        match = re.search(self.PREFIX_PATTERN, value)
        if not match:
            message = (
                f"The source uri {value} is not supported. "
                f"Pattern: {self.PREFIX_PATTERN}"
            )
            self.fail(message, param, ctx)

        return value


@click.command(context_settings=const.CONTEXT_SETTINGS,)
@click.option(
    "-s",
    "--source-uri",
    type=SourceURI(),
    required=True,
    help=(
        "URI of where this data should be downloaded. "
        f"Supported source uri patterns {SourceURI.PREFIX_PATTERN}"
    ),
)
@click.option(
    "-o",
    "--output",
    type=click.Path(exists=True, file_okay=False, writable=True),
    default=const.DEFAULT_DATA_ROOT,
    help="Directory on localhost where datasets should be downloaded.",
)
@click.option(
    "-b",
    "--include-binary",
    is_flag=True,
    default=False,
    help=(
        "Whether to download binary files such as images or LIDAR point "
        "clouds. This flag applies to Datasets where metadata "
        "(e.g. annotation json, dataset catalog, ...) can be separated from "
        "binary files."
    ),
)
@click.option(
    "--access-token",
    type=str,
    default=None,
    help="Unity Simulation access token. "
    "This will override synthetic datasets source-uri for Unity Simulation",
)
@click.option(
    "--checksum-file",
    type=str,
    default=None,
    help="Dataset checksum text file path. "
    "Path can be a HTTP(S) url or a local file path. This will help check the "
    "integrity of the downloaded dataset.",
)
def cli(
    source_uri, output, include_binary, access_token, checksum_file,
):
    """Download datasets to localhost from known locations.

    The download command can support downloading from 3 types of sources

    1. Download from Unity Simulation:

    You can specify project_id, run_execution_id, access_token in source-uri:

    \b
    datasetinsights download \\
        --source-uri=usim://<access_token>@<project_id>/<run_execution_id> \\
        --output=$HOME/data

    Alternatively, you can also override access_token such as:

    \b
    datasetinsights download \\
        --source-uri=usim://<project_id>/<run_execution_id> \\
        --output=$HOME/data \\
        --access-token=<access_token>

    2. Downloading from a public http(s) url:

    \b
    datasetinsights download \\
        --source-uri=http://url/to/file.zip \\
        --output=$HOME/data

    3. Downloading from a GCS url:

    \b
    datasetinsights download \\
        --source-uri=gs://url/to/file.zip \\
        --output=$HOME/data

    or download all objects under the same directory:

    \b
    datasetinsights download \\
        --source-uri=gs://url/to/directory \\
        --output=$HOME/data
    """
    ctx = click.get_current_context()
    logger.debug(f"Called download command with parameters: {ctx.params}")

    downloader = create_dataset_downloader(
        source_uri=source_uri, access_token=access_token
    )
    downloader.download(
        source_uri=source_uri,
        output=output,
        include_binary=include_binary,
        checksum_file=checksum_file,
    )



import logging

import datasetinsights.constants as const
from datasetinsights.datasets.unity_perception import MetricDefinitions, Metrics
from datasetinsights.datasets.unity_perception.tables import SCHEMA_VERSION

logger = logging.getLogger(__name__)


class RenderedObjectInfo:
    """Rendered Object Info in Captures

    This metric stores common object info captured by a sensor in the simulation
    environment. It can be used to calculate object statistics such as
    object count, object rotation and visible pixels.

    Attributes:
        raw_table (pd.DataFrame): rendered object info stored with a tidy
        pandas dataframe. Columns "label_id", "instance_id", "visible_pixels",
        "capture_id, "label_name".

    Examples:

    .. code-block:: python

        >>> # set the data root path to where data was stored
        >>> data_root = "$HOME/data"
        >>> # use rendered object info definition id
        >>> definition_id = "659c6e36-f9f8-4dd6-9651-4a80e51eabc4"
        >>> roinfo = RenderedObjectInfo(data_root, definition_id)
        #total object count per label dataframe
        >>> roinfo.total_counts()
        label_id label_name count
               1    object1    10
               2    object2    21
        #object count per capture dataframe
        >>> roinfo.per_capture_counts()
        capture_id  count
            qwerty     10
            asdfgh     21
    """

    LABEL = "label_id"
    LABEL_READABLE = "label_name"
    INDEX_COLUMN = "capture_id"
    VALUE_COLUMN = "values"
    COUNT_COLUMN = "count"

    def __init__(
        self,
        data_root=const.DEFAULT_DATA_ROOT,
        version=SCHEMA_VERSION,
        def_id=None,
    ):
        """Initialize RenderedObjectInfo

        Args:
            data_root (str): root directory where the dataset was stored
            version (str): synthetic dataset schema version
            def_id (str): rendered object info definition id
        """
        filtered_metrics = Metrics(data_root, version).filter_metrics(def_id)
        label_mappings = self._read_label_mappings(data_root, version, def_id)
        self.raw_table = self._read_filtered_metrics(
            filtered_metrics, label_mappings
        )

    def num_captures(self):
        """Total number of captures

        Returns:
            integer: Total number of captures
        """
        return self.raw_table[self.INDEX_COLUMN].nunique()

    @staticmethod
    def _read_label_mappings(data_root, version, def_id):
        """Read label_mappings from a metric_definition record.

        Args:
            data_root (str): root directory where the dataset was stored
            version (str): synthetic dataset schema version
            def_id (str): rendered object info definition id

        Returns:
            dict: The mappings of {label_id: label_name}
        """
        definition = MetricDefinitions(data_root, version).get_definition(
            def_id
        )
        name = RenderedObjectInfo.LABEL
        readable_name = RenderedObjectInfo.LABEL_READABLE

        return {d[name]: d[readable_name] for d in definition["spec"]}

    @staticmethod
    def _read_filtered_metrics(filtered_metrics, label_mappings):
        """Read label_mappings from a metric_definition record.

        Args:
            filtered_metrics (pd.DataFrame): A pandas dataframe for metrics
                filtered by definition id.
            label_mappings (dict): the mappings of {label_id: label_name}

        Returns:
            pd.DataFrame: rendered object info stored with a tidy
            pandas dataframe. Columns "label_id", "instance_id",
            "visible_pixels", "capture_id, "label_name".
        """
        filtered_metrics[RenderedObjectInfo.LABEL_READABLE] = filtered_metrics[
            RenderedObjectInfo.LABEL
        ].map(label_mappings)
        # Remove metrics data not defined in label_mappings
        filtered_metrics.dropna(
            subset=[RenderedObjectInfo.LABEL_READABLE], inplace=True
        )

        return filtered_metrics

    def total_counts(self):
        """Aggregate Total Object Counts Per Label

        Returns:
            pd.DataFrame: Total object counts table.
                Columns "label_id", "label_name", "count"
        """
        agg = (
            self.raw_table.groupby([self.LABEL, self.LABEL_READABLE])
            .size()
            .to_frame(name=self.COUNT_COLUMN)
            .reset_index()
        )

        return agg

    def per_capture_counts(self):
        """ Aggregate Object Counts Per Label

        Returns:
            pd.DataFrame: Total object counts table.
                Columns "capture_id", "count"
        """
        agg = (
            self.raw_table.groupby(self.INDEX_COLUMN)
            .size()
            .to_frame(name=self.COUNT_COLUMN)
            .reset_index()
        )

        return agg



from .statistics import RenderedObjectInfo
from .visualization.plots import (
    bar_plot,
    grid_plot,
    histogram_plot,
    model_performance_box_plot,
    model_performance_comparison_box_plot,
    plot_bboxes,
    plot_keypoints,
    rotation_plot,
)

__all__ = [
    "bar_plot",
    "grid_plot",
    "histogram_plot",
    "plot_bboxes",
    "model_performance_box_plot",
    "model_performance_comparison_box_plot",
    "rotation_plot",
    "RenderedObjectInfo",
    "plot_keypoints",
]



from .base import create_dataset_downloader
from .gcs_downloader import GCSDatasetDownloader
from .http_downloader import HTTPDatasetDownloader
from .unity_simulation import UnitySimulationDownloader

__all__ = [
    "UnitySimulationDownloader",
    "HTTPDatasetDownloader",
    "create_dataset_downloader",
    "GCSDatasetDownloader",
]
